# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright (c) OpenMMLab. All rights reserved.
# Refer to: https://github.com/open-mmlab/mmaction2/blob/master/mmaction/models/skeleton_gcn/utils/graph.py
import numpy as np


def get_hop_distance(num_node, edge, max_hop=1):
    adj_mat = np.zeros((num_node, num_node))
    for i, j in edge:
        adj_mat[i, j] = 1
        adj_mat[j, i] = 1

    # compute hop steps
    hop_dis = np.zeros((num_node, num_node)) + np.inf
    transfer_mat = [
        np.linalg.matrix_power(adj_mat, d) for d in range(max_hop + 1)
    ]
    arrive_mat = (np.stack(transfer_mat) > 0)
    for d in range(max_hop, -1, -1):
        hop_dis[arrive_mat[d]] = d
    return hop_dis


def normalize_digraph(adj_matrix):
    Dl = np.sum(adj_matrix, 0)
    num_nodes = adj_matrix.shape[0]
    Dn = np.zeros((num_nodes, num_nodes))
    for i in range(num_nodes):
        if Dl[i] > 0:
            Dn[i, i] = Dl[i]**(-1)
    norm_matrix = np.dot(adj_matrix, Dn)
    return norm_matrix


def edge2mat(link, num_node):
    A = np.zeros((num_node, num_node))
    for i, j in link:
        A[j, i] = 1
    return A


class Graph:
    """The Graph to model the skeletons extracted by the openpose.

    Args:
        layout (str): must be one of the following candidates
        - openpose: 18 or 25 joints. For more information, please refer to:
            https://github.com/CMU-Perceptual-Computing-Lab/openpose#output
        - ntu-rgb+d: Is consists of 25 joints. For more information, please
            refer to https://github.com/shahroudy/NTURGB-D

        strategy (str): must be one of the follow candidates
        - uniform: Uniform Labeling
        - distance: Distance Partitioning
        - spatial: Spatial Configuration
        For more information, please refer to the section 'Partition
        Strategies' in our paper (https://arxiv.org/abs/1801.07455).

        max_hop (int): the maximal distance between two connected nodes.
            Default: 1
        dilation (int): controls the spacing between the kernel points.
            Default: 1
    """

    def __init__(self,
                 layout='openpose-18',
                 strategy='uniform',
                 max_hop=1,
                 dilation=1):
        self.max_hop = max_hop
        self.dilation = dilation

        assert layout in [
            'openpose-18', 'openpose-25', 'ntu-rgb+d', 'ntu_edge', 'coco'
        ]
        assert strategy in ['uniform', 'distance', 'spatial', 'agcn']
        self.get_edge(layout)
        self.hop_dis = get_hop_distance(
            self.num_node, self.edge, max_hop=max_hop)
        self.get_adjacency(strategy)

    def __str__(self):
        return self.A

    def get_edge(self, layout):
        """This method returns the edge pairs of the layout."""

        if layout == 'openpose-18':
            self.num_node = 18
            self_link = [(i, i) for i in range(self.num_node)]
            neighbor_link = [(4, 3), (3, 2), (7, 6), (6, 5),
                             (13, 12), (12, 11), (10, 9), (9, 8), (11, 5),
                             (8, 2), (5, 1), (2, 1), (0, 1), (15, 0), (14, 0),
                             (17, 15), (16, 14)]
            self.edge = self_link + neighbor_link
            self.center = 1
        elif layout == 'openpose-25':
            self.num_node = 25
            self_link = [(i, i) for i in range(self.num_node)]
            neighbor_link = [(4, 3), (3, 2), (7, 6), (6, 5), (23, 22),
                             (22, 11), (24, 11), (11, 10), (10, 9), (9, 8),
                             (20, 19), (19, 14), (21, 14), (14, 13), (13, 12),
                             (12, 8), (8, 1), (5, 1), (2, 1), (0, 1), (15, 0),
                             (16, 0), (17, 15), (18, 16)]
            self.self_link = self_link
            self.neighbor_link = neighbor_link
            self.edge = self_link + neighbor_link
            self.center = 1
        elif layout == 'ntu-rgb+d':
            self.num_node = 25
            self_link = [(i, i) for i in range(self.num_node)]
            neighbor_1base = [(1, 2), (2, 21), (3, 21),
                              (4, 3), (5, 21), (6, 5), (7, 6), (8, 7), (9, 21),
                              (10, 9), (11, 10), (12, 11), (13, 1), (14, 13),
                              (15, 14), (16, 15), (17, 1), (18, 17), (19, 18),
                              (20, 19), (22, 23), (23, 8), (24, 25), (25, 12)]
            neighbor_link = [(i - 1, j - 1) for (i, j) in neighbor_1base]
            self.self_link = self_link
            self.neighbor_link = neighbor_link
            self.edge = self_link + neighbor_link
            self.center = 21 - 1
        elif layout == 'ntu_edge':
            self.num_node = 24
            self_link = [(i, i) for i in range(self.num_node)]
            neighbor_1base = [(1, 2), (3, 2), (4, 3), (5, 2), (6, 5), (7, 6),
                              (8, 7), (9, 2), (10, 9), (11, 10), (12, 11),
                              (13, 1), (14, 13), (15, 14), (16, 15), (17, 1),
                              (18, 17), (19, 18), (20, 19), (21, 22), (22, 8),
                              (23, 24), (24, 12)]
            neighbor_link = [(i - 1, j - 1) for (i, j) in neighbor_1base]
            self.edge = self_link + neighbor_link
            self.center = 2
        elif layout == 'coco':
            self.num_node = 17
            self_link = [(i, i) for i in range(self.num_node)]
            neighbor_1base = [[16, 14], [14, 12], [17, 15], [15, 13], [12, 13],
                              [6, 12], [7, 13], [6, 7], [8, 6], [9, 7],
                              [10, 8], [11, 9], [2, 3], [2, 1], [3, 1], [4, 2],
                              [5, 3], [4, 6], [5, 7]]
            neighbor_link = [(i - 1, j - 1) for (i, j) in neighbor_1base]
            self.edge = self_link + neighbor_link
            self.center = 0
        else:
            raise ValueError(f'{layout} is not supported.')

    def get_adjacency(self, strategy):
        """This method returns the adjacency matrix according to strategy."""

        valid_hop = range(0, self.max_hop + 1, self.dilation)
        adjacency = np.zeros((self.num_node, self.num_node))
        for hop in valid_hop:
            adjacency[self.hop_dis == hop] = 1
        normalize_adjacency = normalize_digraph(adjacency)

        if strategy == 'uniform':
            A = np.zeros((1, self.num_node, self.num_node))
            A[0] = normalize_adjacency
            self.A = A
        elif strategy == 'distance':
            A = np.zeros((len(valid_hop), self.num_node, self.num_node))
            for i, hop in enumerate(valid_hop):
                A[i][self.hop_dis == hop] = normalize_adjacency[self.hop_dis ==
                                                                hop]
            self.A = A
        elif strategy == 'spatial':
            A = []
            for hop in valid_hop:
                a_root = np.zeros((self.num_node, self.num_node))
                a_close = np.zeros((self.num_node, self.num_node))
                a_further = np.zeros((self.num_node, self.num_node))
                for i in range(self.num_node):
                    for j in range(self.num_node):
                        if self.hop_dis[j, i] == hop:
                            if self.hop_dis[j, self.center] == self.hop_dis[
                                    i, self.center]:
                                a_root[j, i] = normalize_adjacency[j, i]
                            elif self.hop_dis[j, self.center] > self.hop_dis[
                                    i, self.center]:
                                a_close[j, i] = normalize_adjacency[j, i]
                            else:
                                a_further[j, i] = normalize_adjacency[j, i]
                if hop == 0:
                    A.append(a_root)
                else:
                    A.append(a_root + a_close)
                    A.append(a_further)
            A = np.stack(A)
            self.A = A
        elif strategy == 'agcn':
            A = []
            link_mat = edge2mat(self.self_link, self.num_node)
            In = normalize_digraph(edge2mat(self.neighbor_link, self.num_node))
            outward = [(j, i) for (i, j) in self.neighbor_link]
            Out = normalize_digraph(edge2mat(outward, self.num_node))
            A = np.stack((link_mat, In, Out))
            self.A = A
        else:
            raise ValueError('Do Not Exist This Strategy')



# This is a TensorRT Plugin Python Wrapper Link implementation, original plugin documents refers to
# https://github.com/NVIDIA/TensorRT/tree/main/plugin/
# We use python wrapper to build ONNX-TRTPlugin Engine and then wrapper as a jit script module, after this,
# we could replace some original model's OP with this plugin during Blade Export to speed up those are not
# well optimized by original Blade
# Here we provide a TRTPlugin-EfficientNMS implementation

import torch
from torch import nn


class TRT8_NMS(torch.autograd.Function):
    '''TensorRT NMS operation'''

    @staticmethod
    def forward(
        ctx,
        boxes,
        scores,
        background_class=-1,
        box_coding=1,
        iou_threshold=0.45,
        max_output_boxes=100,
        plugin_version='1',
        score_activation=0,
        score_threshold=0.25,
    ):
        batch_size, num_boxes, num_classes = scores.shape
        num_det = torch.randint(
            0, max_output_boxes, (batch_size, 1), dtype=torch.int32)
        det_boxes = torch.randn(batch_size, max_output_boxes, 4)
        det_scores = torch.randn(batch_size, max_output_boxes)
        det_classes = torch.randint(
            0, num_classes, (batch_size, max_output_boxes), dtype=torch.int32)
        return num_det, det_boxes, det_scores, det_classes

    @staticmethod
    def symbolic(g,
                 boxes,
                 scores,
                 background_class=-1,
                 box_coding=1,
                 iou_threshold=0.45,
                 max_output_boxes=100,
                 plugin_version='1',
                 score_activation=0,
                 score_threshold=0.25):
        out = g.op(
            'TRT::EfficientNMS_TRT',
            boxes,
            scores,
            background_class_i=background_class,
            box_coding_i=box_coding,
            iou_threshold_f=iou_threshold,
            max_output_boxes_i=max_output_boxes,
            plugin_version_s=plugin_version,
            score_activation_i=score_activation,
            score_threshold_f=score_threshold,
            outputs=4)
        nums, boxes, scores, classes = out
        return nums, boxes, scores, classes


class ONNX_TRT8(nn.Module):
    '''onnx module with TensorRT NMS operation.'''

    def __init__(self,
                 max_obj=100,
                 iou_thres=0.45,
                 score_thres=0.25,
                 max_wh=None,
                 device=None):
        super().__init__()
        assert max_wh is None
        self.device = device if device else torch.device('cpu')
        self.background_class = -1,
        self.box_coding = 1,
        self.iou_threshold = iou_thres
        self.max_obj = max_obj
        self.plugin_version = '1'
        self.score_activation = 0
        self.score_threshold = score_thres

    def forward(self, x):
        box = x[:, :, :4]
        conf = x[:, :, 4:5]
        score = x[:, :, 5:]
        score *= conf
        num_det, det_boxes, det_scores, det_classes = TRT8_NMS.apply(
            box, score, self.background_class, self.box_coding,
            self.iou_threshold, self.max_obj, self.plugin_version,
            self.score_activation, self.score_threshold)
        return num_det, det_boxes, det_scores, det_classes


def create_tensorrt_efficientnms(example_scores,
                                 iou_thres=0.45,
                                 score_thres=0.25):
    """

    """
    from torch_blade import tensorrt
    import torch_blade._torch_blade._backends as backends
    import io

    model = torch.jit.trace(
        ONNX_TRT8(iou_thres=iou_thres, score_thres=score_thres),
        example_scores)
    example_outputs = model(example_scores)

    input_names = ['input']
    output_names = [
        'num_det', 'detection_boxes', 'detection_scores', 'detection_classes'
    ]
    with io.BytesIO() as onnx_proto_f:
        torch.onnx.export(
            model,
            example_scores,
            onnx_proto_f,
            input_names=input_names,
            output_names=output_names,
            example_outputs=example_outputs)
        onnx_proto = onnx_proto_f.getvalue()

    def _copy_meta(data, name, dtype, sizes):
        data.name = name
        if dtype.is_floating_point:
            data.dtype = 'Float'
        else:
            data.dtype = 'Int'
        data.sizes = sizes
        return data

    state = backends.EngineState()
    state.inputs = [
        _copy_meta(backends.TensorInfo(), name, tensor.dtype,
                   list(tensor.shape))
        for name, tensor in zip(input_names, [example_scores])
    ]
    state.outputs = [
        _copy_meta(backends.TensorInfo(), name, tensor.dtype, [])
        for name, tensor in zip(output_names, example_outputs)
    ]
    state = tensorrt.cvt_onnx_to_tensorrt(onnx_proto, state, [], dict())

    class Model(torch.nn.Module):

        def __init__(self, state):
            super().__init__()
            self._trt_engine_ext = backends.create_engine(state)

        def forward(self, x):
            return self._trt_engine_ext.execute([x])

    trt_ext = torch.jit.script(Model(state))
    return trt_ext



_base_ = './resnet50_b32x8_100e_jpg.py'

# dataset settings
img_norm_cfg = dict(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
mean = [x * 255 for x in img_norm_cfg['mean']]
std = [x * 255 for x in img_norm_cfg['std']]
size = 224
train_pipeline = [
    dict(type='DaliImageDecoder'),
    dict(type='DaliRandomResizedCrop', size=size, random_area=(0.08, 1.0)),
    dict(
        type='DaliCropMirrorNormalize',
        crop=[size, size],
        mean=mean,
        std=std,
        crop_pos_x=[0.0, 1.0],
        crop_pos_y=[0.0, 1.0],
        prob=0.5)
]
val_pipeline = [
    dict(type='DaliImageDecoder'),
    dict(type='DaliResize', resize_shorter=size * 1.15),
    dict(
        type='DaliCropMirrorNormalize',
        crop=[size, size],
        mean=mean,
        std=std,
        prob=0.0)
]
data = dict(
    imgs_per_gpu=32,  # total 256
    workers_per_gpu=4,
    train=dict(
        type='DaliImageNetTFRecordDataSet',
        data_source=dict(
            type='ClsSourceImageNetTFRecord',
            file_pattern='oss://path/to/data/imagenet-tfrecord/train-*',
            # root=data_root,  # pick one of `file_pattern` and `root&list_file`
            # list_file=data_root + 'meta/train.txt',
            cache_path='data/cache/train',
        ),
        pipeline=train_pipeline,
        label_offset=1),
    val=dict(
        imgs_per_gpu=50,
        type='DaliImageNetTFRecordDataSet',
        data_source=dict(
            type='ClsSourceImageNetTFRecord',
            file_pattern='oss://path/to/data/imagenet-tfrecord/train-*',
            # root=data_root,
            # list_file=data_root + 'meta/train_relative.txt',
            cache_path='data/cache/val',
        ),
        pipeline=val_pipeline,
        random_shuffle=False,
        label_offset=1))



# Copyright (c) Alibaba, Inc. and its affiliates.
from .encoder_decoder import EncoderDecoder
from .heads import *
from .mask2former import Mask2Former
from .sampler import BasePixelSampler, OHEMPixelSampler, build_pixel_sampler



import math
from typing import Optional

import torch
from torch import Tensor, nn
from torch.nn import functional as F

from easycv.framework.errors import RuntimeError, ValueError


class PositionEmbeddingSine(nn.Module):
    """
    This is a more standard version of the position embedding, very similar to the one
    used by the Attention is all you need paper, generalized to work on images.
    """

    def __init__(self,
                 num_pos_feats=64,
                 temperature=10000,
                 normalize=False,
                 scale=None):
        super().__init__()
        self.num_pos_feats = num_pos_feats
        self.temperature = temperature
        self.normalize = normalize
        if scale is not None and normalize is False:
            raise ValueError('normalize should be True if scale is passed')
        if scale is None:
            scale = 2 * math.pi
        self.scale = scale

    def forward(self, x, mask=None):
        if mask is None:
            mask = torch.zeros((x.size(0), x.size(2), x.size(3)),
                               device=x.device,
                               dtype=torch.bool)
        not_mask = ~mask
        y_embed = not_mask.cumsum(1, dtype=torch.float32)
        x_embed = not_mask.cumsum(2, dtype=torch.float32)
        if self.normalize:
            eps = 1e-6
            y_embed = y_embed / (y_embed[:, -1:, :] + eps) * self.scale
            x_embed = x_embed / (x_embed[:, :, -1:] + eps) * self.scale

        dim_t = torch.arange(
            self.num_pos_feats, dtype=torch.float32, device=x.device)
        dim_t = self.temperature**(2 * (dim_t // 2) / self.num_pos_feats)

        pos_x = x_embed[:, :, :, None] / dim_t
        pos_y = y_embed[:, :, :, None] / dim_t
        pos_x = torch.stack(
            (pos_x[:, :, :, 0::2].sin(), pos_x[:, :, :, 1::2].cos()),
            dim=4).flatten(3)
        pos_y = torch.stack(
            (pos_y[:, :, :, 0::2].sin(), pos_y[:, :, :, 1::2].cos()),
            dim=4).flatten(3)
        pos = torch.cat((pos_y, pos_x), dim=3).permute(0, 3, 1, 2)
        return pos

    def __repr__(self, _repr_indent=4):
        head = 'Positional encoding ' + self.__class__.__name__
        body = [
            'num_pos_feats: {}'.format(self.num_pos_feats),
            'temperature: {}'.format(self.temperature),
            'normalize: {}'.format(self.normalize),
            'scale: {}'.format(self.scale),
        ]
        # _repr_indent = 4
        lines = [head] + [' ' * _repr_indent + line for line in body]
        return '\n'.join(lines)


def _get_activation_fn(activation):
    """Return an activation function given a string"""
    if activation == 'relu':
        return F.relu
    if activation == 'gelu':
        return F.gelu
    if activation == 'glu':
        return F.glu
    raise RuntimeError(F'activation should be relu/gelu, not {activation}.')


class SelfAttentionLayer(nn.Module):

    def __init__(self,
                 d_model,
                 nhead,
                 dropout=0.0,
                 activation='relu',
                 normalize_before=False):
        super().__init__()
        self.self_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)

        self.norm = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

        self.activation = _get_activation_fn(activation)
        self.normalize_before = normalize_before

        self._reset_parameters()

    def _reset_parameters(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def with_pos_embed(self, tensor, pos: Optional[Tensor]):
        return tensor if pos is None else tensor + pos

    def forward_post(self,
                     tgt,
                     tgt_mask: Optional[Tensor] = None,
                     tgt_key_padding_mask: Optional[Tensor] = None,
                     query_pos: Optional[Tensor] = None):
        q = k = self.with_pos_embed(tgt, query_pos)
        tgt2 = self.self_attn(
            q,
            k,
            value=tgt,
            attn_mask=tgt_mask,
            key_padding_mask=tgt_key_padding_mask)[0]
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm(tgt)

        return tgt

    def forward_pre(self,
                    tgt,
                    tgt_mask: Optional[Tensor] = None,
                    tgt_key_padding_mask: Optional[Tensor] = None,
                    query_pos: Optional[Tensor] = None):
        tgt2 = self.norm(tgt)
        q = k = self.with_pos_embed(tgt2, query_pos)
        tgt2 = self.self_attn(
            q,
            k,
            value=tgt2,
            attn_mask=tgt_mask,
            key_padding_mask=tgt_key_padding_mask)[0]
        tgt = tgt + self.dropout(tgt2)

        return tgt

    def forward(self,
                tgt,
                tgt_mask: Optional[Tensor] = None,
                tgt_key_padding_mask: Optional[Tensor] = None,
                query_pos: Optional[Tensor] = None):
        if self.normalize_before:
            return self.forward_pre(tgt, tgt_mask, tgt_key_padding_mask,
                                    query_pos)
        return self.forward_post(tgt, tgt_mask, tgt_key_padding_mask,
                                 query_pos)


class CrossAttentionLayer(nn.Module):

    def __init__(self,
                 d_model,
                 nhead,
                 dropout=0.0,
                 activation='relu',
                 normalize_before=False):
        super().__init__()
        self.multihead_attn = nn.MultiheadAttention(
            d_model, nhead, dropout=dropout)

        self.norm = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

        self.activation = _get_activation_fn(activation)
        self.normalize_before = normalize_before

        self._reset_parameters()

    def _reset_parameters(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def with_pos_embed(self, tensor, pos: Optional[Tensor]):
        return tensor if pos is None else tensor + pos

    def forward_post(self,
                     tgt,
                     memory,
                     memory_mask: Optional[Tensor] = None,
                     memory_key_padding_mask: Optional[Tensor] = None,
                     pos: Optional[Tensor] = None,
                     query_pos: Optional[Tensor] = None):
        tgt2 = self.multihead_attn(
            query=self.with_pos_embed(tgt, query_pos),
            key=self.with_pos_embed(memory, pos),
            value=memory,
            attn_mask=memory_mask,
            key_padding_mask=memory_key_padding_mask)[0]
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm(tgt)

        return tgt

    def forward_pre(self,
                    tgt,
                    memory,
                    memory_mask: Optional[Tensor] = None,
                    memory_key_padding_mask: Optional[Tensor] = None,
                    pos: Optional[Tensor] = None,
                    query_pos: Optional[Tensor] = None):
        tgt2 = self.norm(tgt)
        tgt2 = self.multihead_attn(
            query=self.with_pos_embed(tgt2, query_pos),
            key=self.with_pos_embed(memory, pos),
            value=memory,
            attn_mask=memory_mask,
            key_padding_mask=memory_key_padding_mask)[0]
        tgt = tgt + self.dropout(tgt2)

        return tgt

    def forward(self,
                tgt,
                memory,
                memory_mask: Optional[Tensor] = None,
                memory_key_padding_mask: Optional[Tensor] = None,
                pos: Optional[Tensor] = None,
                query_pos: Optional[Tensor] = None):
        if self.normalize_before:
            return self.forward_pre(tgt, memory, memory_mask,
                                    memory_key_padding_mask, pos, query_pos)
        return self.forward_post(tgt, memory, memory_mask,
                                 memory_key_padding_mask, pos, query_pos)


class FFNLayer(nn.Module):

    def __init__(self,
                 d_model,
                 dim_feedforward=2048,
                 dropout=0.0,
                 activation='relu',
                 normalize_before=False):
        super().__init__()
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm = nn.LayerNorm(d_model)

        self.activation = _get_activation_fn(activation)
        self.normalize_before = normalize_before

        self._reset_parameters()

    def _reset_parameters(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def with_pos_embed(self, tensor, pos: Optional[Tensor]):
        return tensor if pos is None else tensor + pos

    def forward_post(self, tgt):
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm(tgt)
        return tgt

    def forward_pre(self, tgt):
        tgt2 = self.norm(tgt)
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt2))))
        tgt = tgt + self.dropout(tgt2)
        return tgt

    def forward(self, tgt):
        if self.normalize_before:
            return self.forward_pre(tgt)
        return self.forward_post(tgt)


class MLP(nn.Module):
    """ Very simple multi-layer perceptron (also called FFN)"""

    def __init__(self, input_dim, hidden_dim, output_dim, num_layers):
        super().__init__()
        self.num_layers = num_layers
        h = [hidden_dim] * (num_layers - 1)
        self.layers = nn.ModuleList(
            nn.Linear(n, k) for n, k in zip([input_dim] + h, h + [output_dim]))

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            x = F.relu(layer(x)) if i < self.num_layers - 1 else layer(x)
        return x


class MultiScaleMaskedTransformerDecoder(nn.Module):

    def __init__(
        self,
        in_channels,
        mask_classification=True,
        *,
        num_classes: int = 80,
        hidden_dim: int = 256,
        num_queries: int = 100,
        nheads: int = 8,
        dim_feedforward: int = 2048,
        dec_layers: int = 9,
        pre_norm: bool = False,
        mask_dim: int = 256,
        enforce_input_project: bool = False,
    ):
        """
        NOTE: this interface is experimental.
        Args:
            in_channels: channels of the input features
            mask_classification: whether to add mask classifier or not
            num_classes: number of classes
            hidden_dim: Transformer feature dimension
            num_queries: number of queries
            nheads: number of heads
            dim_feedforward: feature dimension in feedforward network
            enc_layers: number of Transformer encoder layers
            dec_layers: number of Transformer decoder layers
            pre_norm: whether to use pre-LayerNorm or not
            mask_dim: mask feature dimension
            enforce_input_project: add input project 1x1 conv even if input
                channels and hidden dim is identical
        """
        super().__init__()

        assert mask_classification, 'Only support mask classification model'
        self.mask_classification = mask_classification

        # positional encoding
        N_steps = hidden_dim // 2
        self.pe_layer = PositionEmbeddingSine(N_steps, normalize=True)

        # define Transformer decoder here
        self.num_heads = nheads
        self.num_layers = dec_layers
        self.transformer_self_attention_layers = nn.ModuleList()
        self.transformer_cross_attention_layers = nn.ModuleList()
        self.transformer_ffn_layers = nn.ModuleList()

        for _ in range(self.num_layers):
            self.transformer_self_attention_layers.append(
                SelfAttentionLayer(
                    d_model=hidden_dim,
                    nhead=nheads,
                    dropout=0.0,
                    normalize_before=pre_norm,
                ))

            self.transformer_cross_attention_layers.append(
                CrossAttentionLayer(
                    d_model=hidden_dim,
                    nhead=nheads,
                    dropout=0.0,
                    normalize_before=pre_norm,
                ))

            self.transformer_ffn_layers.append(
                FFNLayer(
                    d_model=hidden_dim,
                    dim_feedforward=dim_feedforward,
                    dropout=0.0,
                    normalize_before=pre_norm,
                ))

        self.decoder_norm = nn.LayerNorm(hidden_dim)

        self.num_queries = num_queries
        # learnable query features
        self.query_feat = nn.Embedding(num_queries, hidden_dim)
        # learnable query p.e.
        self.query_embed = nn.Embedding(num_queries, hidden_dim)

        # level embedding (we always use 3 scales)
        self.num_feature_levels = 3
        self.level_embed = nn.Embedding(self.num_feature_levels, hidden_dim)
        self.input_proj = nn.ModuleList()
        for _ in range(self.num_feature_levels):
            if in_channels != hidden_dim or enforce_input_project:
                # self.input_proj.append(Conv2d(in_channels, hidden_dim, kernel_size=1))
                # weight_init.c2_xavier_fill(self.input_proj[-1])
                self.input_proj.append(
                    torch.nn.Conv2d(in_channels, hidden_dim, kernel_size=1))
                nn.init.kaiming_uniform_(self.input_proj[-1].weight, a=1)
                if self.self.input_proj[-1].bias is not None:
                    nn.init.constant(self.self.input_proj[-1].bias, 0)
            else:
                self.input_proj.append(nn.Sequential())

        # output FFNs
        if self.mask_classification:
            self.class_embed = nn.Linear(hidden_dim, num_classes + 1)
        self.mask_embed = MLP(hidden_dim, hidden_dim, mask_dim, 3)

    def forward(self, x, mask_features, mask=None):
        # x is a list of multi-scale feature
        assert len(x) == self.num_feature_levels
        src = []
        pos = []
        size_list = []

        # disable mask, it does not affect performance
        del mask

        for i in range(self.num_feature_levels):
            size_list.append(x[i].shape[-2:])
            pos.append(self.pe_layer(x[i], None).flatten(2))
            src.append(self.input_proj[i](x[i]).flatten(2) +
                       self.level_embed.weight[i][None, :, None])

            # flatten NxCxHxW to HWxNxC
            pos[-1] = pos[-1].permute(2, 0, 1)
            src[-1] = src[-1].permute(2, 0, 1)

        _, bs, _ = src[0].shape

        # QxNxC
        query_embed = self.query_embed.weight.unsqueeze(1).repeat(1, bs, 1)
        output = self.query_feat.weight.unsqueeze(1).repeat(1, bs, 1)

        predictions_class = []
        predictions_mask = []

        # prediction heads on learnable query features
        outputs_class, outputs_mask, attn_mask = self.forward_prediction_heads(
            output, mask_features, attn_mask_target_size=size_list[0])
        predictions_class.append(outputs_class)
        predictions_mask.append(outputs_mask)

        for i in range(self.num_layers):
            level_index = i % self.num_feature_levels
            attn_mask[torch.where(
                attn_mask.sum(-1) == attn_mask.shape[-1])] = False
            # attention: cross-attention first
            output = self.transformer_cross_attention_layers[i](
                output,
                src[level_index],
                memory_mask=attn_mask,
                memory_key_padding_mask=
                None,  # here we do not apply masking on padded region
                pos=pos[level_index],
                query_pos=query_embed)

            output = self.transformer_self_attention_layers[i](
                output,
                tgt_mask=None,
                tgt_key_padding_mask=None,
                query_pos=query_embed)

            # FFN
            output = self.transformer_ffn_layers[i](output)

            outputs_class, outputs_mask, attn_mask = self.forward_prediction_heads(
                output,
                mask_features,
                attn_mask_target_size=size_list[(i + 1) %
                                                self.num_feature_levels])
            predictions_class.append(outputs_class)
            predictions_mask.append(outputs_mask)

        assert len(predictions_class) == self.num_layers + 1

        out = {
            'pred_logits':
            predictions_class[-1],
            'pred_masks':
            predictions_mask[-1],
            'aux_outputs':
            self._set_aux_loss(
                predictions_class if self.mask_classification else None,
                predictions_mask)
        }
        return out

    def forward_prediction_heads(self, output, mask_features,
                                 attn_mask_target_size):
        decoder_output = self.decoder_norm(output)
        decoder_output = decoder_output.transpose(0, 1)
        outputs_class = self.class_embed(decoder_output)
        mask_embed = self.mask_embed(decoder_output)
        outputs_mask = torch.einsum('bqc,bchw->bqhw', mask_embed,
                                    mask_features)

        # NOTE: prediction is of higher-resolution
        # [B, Q, H, W] -> [B, Q, H*W] -> [B, h, Q, H*W] -> [B*h, Q, HW]
        attn_mask = F.interpolate(
            outputs_mask,
            size=attn_mask_target_size,
            mode='bilinear',
            align_corners=False)
        # must use bool type
        # If a BoolTensor is provided, positions with ``True`` are not allowed to attend while ``False`` values will be unchanged.
        attn_mask = (attn_mask.sigmoid().flatten(2).unsqueeze(1).repeat(
            1, self.num_heads, 1, 1).flatten(0, 1) < 0.5).bool()
        attn_mask = attn_mask.detach()

        return outputs_class, outputs_mask, attn_mask

    @torch.jit.unused
    def _set_aux_loss(self, outputs_class, outputs_seg_masks):
        # this is a workaround to make torchscript happy, as torchscript
        # doesn't support dictionary with non-homogeneous values, such
        # as a dict having both a Tensor and a list.
        if self.mask_classification:
            return [{
                'pred_logits': a,
                'pred_masks': b
            } for a, b in zip(outputs_class[:-1], outputs_seg_masks[:-1])]
        else:
            return [{'pred_masks': b} for b in outputs_seg_masks[:-1]]



# Copyright (c) Alibaba, Inc. and its affiliates.
import io

import cv2
import numpy as np
from cv2 import IMREAD_COLOR
from PIL import Image

from easycv import file
from easycv.framework.errors import KeyError, ValueError
from .utils import is_url_path

try:
    from turbojpeg import TurboJPEG, TJCS_RGB, TJPF_BGR
    turbo_jpeg = TurboJPEG()
    turbo_jpeg_mode = {'RGB': TJCS_RGB, 'BGR': TJPF_BGR}
except:
    turbo_jpeg = None
    turbo_jpeg_mode = None


def load_image_with_pillow(content, mode='BGR', dtype=np.uint8):
    with io.BytesIO(content) as buff:
        image = Image.open(buff)

        if mode.upper() != 'BGR':
            if image.mode.upper() != mode.upper():
                image = image.convert(mode.upper())
            img = np.asarray(image, dtype=dtype)
        else:
            if image.mode.upper() != 'RGB':
                image = image.convert('RGB')
            img = np.asarray(image, dtype=dtype)
            img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

    return img


def load_image_with_turbojpeg(content, mode='BGR', dtype=np.uint8):
    assert mode.upper() in turbo_jpeg_mode
    if turbo_jpeg is None or turbo_jpeg_mode is None:
        raise ValueError(
            'Please install turbojpeg by "pip install PyTurboJPEG" !')

    img = turbo_jpeg.decode(
        content, pixel_format=turbo_jpeg_mode[mode.upper()])

    if img.dtype != dtype:
        img = img.astype(dtype)

    return img


def load_image_with_cv2(content, mode='BGR', dtype=np.uint8):
    assert mode.upper() in ['BGR', 'RGB']

    img_np = np.frombuffer(content, np.uint8)
    img = cv2.imdecode(img_np, flags=IMREAD_COLOR)

    if mode.upper() == 'RGB':
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    if img.dtype != dtype:
        img = img.astype(dtype)

    return img


def _load_image(fp, mode='BGR', dtype=np.uint8, backend='pillow'):
    if backend == 'pillow':
        img = load_image_with_pillow(fp, mode=mode, dtype=dtype)
    elif backend == 'turbojpeg':
        img = load_image_with_turbojpeg(fp, mode=mode, dtype=dtype)
    elif backend == 'cv2':
        img = load_image_with_cv2(fp, mode=mode, dtype=dtype)
    else:
        raise KeyError(
            'Only support backend in ["pillow", "turbojpeg", "cv2"]')
    return img


def load_image(img_path, mode='BGR', dtype=np.uint8, backend='pillow'):
    """Load image file, return np.ndarray.

    Args:
        img_path (str): Image file path.
        mode (str): Order of channel, candidates are `bgr` and `rgb`.
        dtype : Output data type.
        backend (str): The image decoding backend type. Options are `cv2`, `pillow`, `turbojpeg`.
    """
    img = None
    if is_url_path(img_path):
        from mmcv.fileio.file_client import HTTPBackend
        client = HTTPBackend()
        img_bytes = client.get(img_path)
        img = _load_image(img_bytes, mode=mode, dtype=dtype, backend=backend)
    else:
        with file.io.open(img_path, 'rb') as infile:
            img = _load_image(
                infile.read(), mode=mode, dtype=dtype, backend=backend)

    return img



# Copyright (c) Alibaba, Inc. and its affiliates.
import copy
import json

import mmcv
import numpy as np
import torch
from mmcv.image import imwrite
from mmcv.utils.path import is_filepath
from mmcv.visualization.image import imshow

from easycv.core.visualization import imshow_bboxes, imshow_keypoints
from easycv.datasets.pose.data_sources.top_down import DatasetInfo
from easycv.datasets.pose.pipelines.transforms import bbox_cs2xyxy
from easycv.file import io
from easycv.predictors.builder import PREDICTORS, build_predictor
from easycv.utils.checkpoint import load_checkpoint
from easycv.utils.config_tools import mmcv_config_fromfile
from easycv.utils.misc import deprecated
from .base import InputProcessor, OutputProcessor, PredictorV2

np.set_printoptions(suppress=True)


def _box2cs(image_size, box):
    """This encodes bbox(x,y,w,h) into (center, scale)
    Args:
        x, y, w, h
    Returns:
        tuple: A tuple containing center and scale.
        - np.ndarray[float32](2,): Center of the bbox (x, y).
        - np.ndarray[float32](2,): Scale of the bbox w & h.
    """

    x, y, w, h = box[:4]
    aspect_ratio = image_size[0] / image_size[1]
    center = np.array([x + w * 0.5, y + h * 0.5], dtype=np.float32)

    if w > aspect_ratio * h:
        h = w * 1.0 / aspect_ratio
    elif w < aspect_ratio * h:
        w = h * aspect_ratio

    # pixel std is 200.0
    scale = np.array([w / 200.0, h / 200.0], dtype=np.float32)
    scale = scale * 1.25

    return center, scale


def vis_pose_result(
    model,
    img,
    result,
    radius=4,
    thickness=1,
    kpt_score_thr=0.3,
    bbox_color='green',
    dataset_info=None,
    out_file=None,
    pose_kpt_color=None,
    pose_link_color=None,
    text_color='white',
    font_scale=0.5,
    bbox_thickness=1,
    win_name='',
    show=False,
    wait_time=0,
):
    """Visualize the detection results on the image.

    Args:
        model (nn.Module): The loaded detector.
        img (str | np.ndarray): Image filename or loaded image.
        result (list[dict]): The results to draw over `img`
                (bbox_result, pose_result).
        radius (int): Radius of circles.
        thickness (int): Thickness of lines.
        kpt_score_thr (float): The threshold to visualize the keypoints.
        skeleton (list[tuple()]): Default None.
        out_file (str|None): The filename of the output visualization image.
        show (bool): Whether to show the image. Default: False.
        wait_time (int): Value of waitKey param.
            Default: 0.
        out_file (str or None): The filename to write the image.
            Default: None.
    """

    # get dataset info
    if (dataset_info is None and hasattr(model, 'cfg')
            and 'dataset_info' in model.cfg):
        dataset_info = DatasetInfo(model.cfg.dataset_info)

    if not dataset_info:
        raise ValueError('Please provide `dataset_info`!')

    skeleton = dataset_info.skeleton
    pose_kpt_color = dataset_info.pose_kpt_color
    pose_link_color = dataset_info.pose_link_color

    if hasattr(model, 'module'):
        model = model.module

    img = mmcv.imread(img)
    img = img.copy()

    bbox_result = result.get('bbox', [])
    pose_result = result['keypoints']

    if len(bbox_result) > 0:
        bboxes = np.vstack(bbox_result)
        labels = None
        if 'label' in result:
            labels = result['label']
        # draw bounding boxes
        imshow_bboxes(
            img,
            bboxes,
            labels=labels,
            colors=bbox_color,
            text_color=text_color,
            thickness=bbox_thickness,
            font_scale=font_scale,
            show=False)

    imshow_keypoints(img, pose_result, skeleton, kpt_score_thr, pose_kpt_color,
                     pose_link_color, radius, thickness)

    if show:
        imshow(img, win_name, wait_time)
    if out_file is not None:
        imwrite(img, out_file)

    return img


class PoseTopDownInputProcessor(InputProcessor):

    def __init__(self,
                 cfg,
                 dataset_info,
                 detection_predictor_config,
                 bbox_thr=None,
                 pipelines=None,
                 batch_size=1,
                 cat_id=None,
                 mode='BGR'):
        self.detection_predictor = build_predictor(detection_predictor_config)
        self.dataset_info = dataset_info
        self.bbox_thr = bbox_thr
        self.cat_id = cat_id
        super().__init__(
            cfg,
            pipelines=pipelines,
            batch_size=batch_size,
            threads=1,
            mode=mode)

    def get_detection_outputs(self, input, cat_id=None):
        det_results = self.detection_predictor(input['img'], keep_inputs=False)
        person_results = self._process_detection_results(
            det_results, cat_id=cat_id)
        return person_results

    def _process_detection_results(self, det_results, cat_id=None):
        """Process det results, and return a list of bboxes.

        Args:
            det_results (list|tuple): det results.
            cat_id (int | str): category id or name to reserve, if None, reserve all detection results.

        Returns:
            person_results (list): a list of detected bounding boxes
        """
        # Only support one sample/image
        if isinstance(det_results, tuple):
            det_results = det_results[0]
        elif isinstance(det_results, list):
            det_results = det_results[0]
        else:
            det_results = det_results

        bboxes = det_results['detection_boxes']
        scores = det_results['detection_scores']
        classes = det_results['detection_classes']

        if cat_id is not None:
            if isinstance(cat_id, str):
                assert cat_id in self.detection_predictor.cfg.CLASSES, f'cat_id "{cat_id}" not in detection classes list: {self.detection_predictor.cfg.CLASSES}'
                assert det_results.get('detection_class_names',
                                       None) is not None
                detection_class_names = det_results['detection_class_names']
                keeped_ids = [
                    i for i in range(len(detection_class_names))
                    if str(detection_class_names[i]) == str(cat_id)
                ]
            else:
                keeped_ids = classes == cat_id
            bboxes = bboxes[keeped_ids]
            scores = scores[keeped_ids]
            classes = classes[keeped_ids]

        person_results = []
        for idx, bbox in enumerate(bboxes):
            person = {}
            bbox = np.append(bbox, scores[idx])
            person['bbox'] = bbox
            person_results.append(person)

        return person_results

    def process_single(self, input):
        output = super()._load_input(input)

        person_results = self.get_detection_outputs(output, cat_id=self.cat_id)

        box_id = 0

        # Select bboxes by score threshold
        bboxes = np.array([res['bbox'] for res in person_results])
        if self.bbox_thr is not None:
            assert bboxes.shape[1] == 5
            valid_idx = np.where(bboxes[:, 4] > self.bbox_thr)[0]
            bboxes = bboxes[valid_idx]
            person_results = [person_results[i] for i in valid_idx]

        results = []
        for person_result in person_results:
            box = person_result['bbox']  # x,y,x,y,s
            boxc = [box[0], box[1], box[2] - box[0],
                    box[3] - box[1]]  # x,y,w,h
            center, scale = _box2cs(self.cfg.data_cfg['image_size'], boxc)
            data = {
                'image_id':
                0,
                'center':
                center,
                'scale':
                scale,
                'bbox':
                box,
                'bbox_score':
                box[4] if len(box) == 5 else 1,
                'bbox_id':
                box_id,  # need to be assigned if batch_size > 1
                'joints_3d':
                np.zeros((self.cfg.data_cfg.num_joints, 3), dtype=np.float32),
                'joints_3d_visible':
                np.zeros((self.cfg.data_cfg.num_joints, 3), dtype=np.float32),
                'rotation':
                0,
                'flip_pairs':
                self.dataset_info.flip_pairs,
                'ann_info': {
                    'image_size': np.array(self.cfg.data_cfg['image_size']),
                    'num_joints': self.cfg.data_cfg['num_joints'],
                },
                'image_file':
                output['filename'],
                'img':
                output['img'],
                'img_shape':
                output['img_shape'],
                'ori_shape':
                output['ori_shape'],
                'img_fields':
                output['img_fields'],
            }
            box_id += 1
            data_processor = self.processor(data)
            data_processor['bbox'] = box
            results.append(data_processor)

        return results

    def __call__(self, inputs):
        """Process all inputs list. And collate to batch and put to target device.
        If you need custom ops to load or process a batch samples, you need to reimplement it.
        """
        batch_outputs = []
        for inp in inputs:
            for res in self.process_single(inp):
                batch_outputs.append(res)

        if len(batch_outputs) < 1:
            return batch_outputs

        batch_outputs = self._collate_fn(batch_outputs)
        batch_outputs['img_metas']._data = [[
            img_meta[i] for img_meta in batch_outputs['img_metas']._data
            for i in range(len(img_meta))
        ]]
        return batch_outputs


class PoseTopDownOutputProcessor(OutputProcessor):

    def __call__(self, inputs):
        output = {}
        output['keypoints'] = inputs['preds']
        output['bbox'] = np.array(inputs['boxes'])  # x1, y1, x2, y2 score

        return output


# TODO: Fix when multi people are detected in each sample,
# all the people results will be passed to the pose model,
# resulting in a dynamic batch_size, which is not supported by jit script model.
@PREDICTORS.register_module()
class PoseTopDownPredictor(PredictorV2):
    """Pose topdown predictor.
        Args:
            model_path (str): Path of model path.
            config_file (Optinal[str]): Config file path for model and processor to init. Defaults to None.
            detection_model_config: Dict of person detection model predictor config,
                example like ``dict(type="", model_path="", config_file="", ......)``
            batch_size (int): Batch size for forward.
            bbox_thr (float): Bounding box threshold to filter output results of detection model
            cat_id (int | str): Category id or name to filter target objects.
            device (str | torch.device): Support str('cuda' or 'cpu') or torch.device, if is None, detect device automatically.
            save_results (bool): Whether to save predict results.
            save_path (str): File path for saving results, only valid when `save_results` is True.
            pipelines (list[dict]): Data pipeline configs.
            mode (str): The image mode into the model.
    """

    def __init__(self,
                 model_path,
                 config_file=None,
                 detection_predictor_config=None,
                 batch_size=1,
                 bbox_thr=None,
                 cat_id=None,
                 device=None,
                 pipelines=None,
                 save_results=False,
                 save_path=None,
                 mode='BGR',
                 model_type=None,
                 *args,
                 **kwargs):
        assert batch_size == 1, 'Only support batch_size=1 now!'
        self.cat_id = cat_id
        self.bbox_thr = bbox_thr
        self.detection_predictor_config = detection_predictor_config

        self.model_type = model_type
        if self.model_type is None:
            if model_path.endswith('jit'):
                assert config_file is not None
                self.model_type = 'jit'
            elif model_path.endswith('blade'):
                import torch_blade
                assert config_file is not None
                self.model_type = 'blade'
            else:
                self.model_type = 'raw'
        assert self.model_type in ['raw', 'jit', 'blade']

        super(PoseTopDownPredictor, self).__init__(
            model_path,
            config_file=config_file,
            batch_size=batch_size,
            device=device,
            save_results=save_results,
            save_path=save_path,
            pipelines=pipelines,
            input_processor_threads=1,
            mode=mode,
            *args,
            **kwargs)
        if hasattr(self.cfg, 'dataset_info'):
            dataset_info = self.cfg.dataset_info
            if is_filepath(dataset_info):
                cfg = mmcv_config_fromfile(dataset_info)
                dataset_info = cfg._cfg_dict['dataset_info']
        else:
            from easycv.datasets.pose.data_sources.coco import COCO_DATASET_INFO
            dataset_info = COCO_DATASET_INFO

        self.dataset_info = DatasetInfo(dataset_info)

    def _build_model(self):
        if self.model_type != 'raw':
            with io.open(self.model_path, 'rb') as infile:
                model = torch.jit.load(infile, self.device)
        else:
            model = super()._build_model()
        return model

    def prepare_model(self):
        """Build model from config file by default.
        If the model is not loaded from a configuration file, e.g. torch jit model, you need to reimplement it.
        """
        model = self._build_model()
        model.to(self.device)
        model.eval()
        if self.model_type == 'raw':
            load_checkpoint(model, self.model_path, map_location='cpu')
        return model

    def model_forward(self, inputs, return_heatmap=False):
        boxes = inputs['bbox'].cpu().numpy()
        if self.model_type == 'raw':
            with torch.no_grad():
                result = self.model(
                    **inputs, mode='test', return_heatmap=return_heatmap)
        else:
            img_metas = inputs['img_metas']
            with torch.no_grad():
                img = inputs['img'].to(self.device)
                tensor_img_metas = copy.deepcopy(img_metas)
                for meta in tensor_img_metas:
                    meta.pop('image_file')
                    for k, v in meta.items():
                        meta[k] = torch.tensor(v)
                output_heatmap = self.model(img, tensor_img_metas)

            from easycv.models.pose.heads.topdown_heatmap_base_head import decode_heatmap
            output_heatmap = output_heatmap.cpu().numpy()
            result = decode_heatmap(output_heatmap, img_metas,
                                    self.cfg.model.test_cfg)

        result['boxes'] = np.array(boxes)
        return result

    def get_input_processor(self):
        return PoseTopDownInputProcessor(
            cfg=self.cfg,
            dataset_info=self.dataset_info,
            detection_predictor_config=self.detection_predictor_config,
            bbox_thr=self.bbox_thr,
            pipelines=self.pipelines,
            batch_size=self.batch_size,
            cat_id=self.cat_id,
            mode=self.mode)

    def get_output_processor(self):
        return PoseTopDownOutputProcessor()

    def show_result(self,
                    image,
                    keypoints,
                    radius=4,
                    thickness=3,
                    kpt_score_thr=0.3,
                    bbox_color='green',
                    show=False,
                    save_path=None):
        vis_result = vis_pose_result(
            self.model,
            image,
            keypoints,
            dataset_info=self.dataset_info,
            radius=radius,
            thickness=thickness,
            kpt_score_thr=kpt_score_thr,
            bbox_color=bbox_color,
            show=show,
            out_file=save_path)

        return vis_result


class _TorchPoseTopDownOutputProcessor(PoseTopDownOutputProcessor):

    def __call__(self, inputs):
        output = super(_TorchPoseTopDownOutputProcessor, self).__call__(inputs)

        bbox = output['bbox']
        keypoints = output['keypoints']
        results = []
        for i in range(len(keypoints)):
            results.append({'bbox': bbox[i], 'keypoints': keypoints[i]})
        return {'pose_results': results}


@deprecated(reason='Please use PoseTopDownPredictor.')
@PREDICTORS.register_module()
class TorchPoseTopDownPredictorWithDetector(PoseTopDownPredictor):

    def __init__(
        self,
        model_path,
        model_config={
            'pose': {
                'bbox_thr': 0.3,
                'format': 'xywh'
            },
            'detection': {
                'model_type': None,
                'reserved_classes': [],
                'score_thresh': 0.0,
            }
        },
    ):
        """
        init model

        Args:
          model_path: pose and detection model file path, split with `,`,
                      make sure the first is pose model, second is detection model
          model_config: config string for model to init, in json format
        """
        if isinstance(model_config, str):
            model_config = json.loads(model_config)

        reserved_classes = model_config['detection'].pop(
            'reserved_classes', [])
        if len(reserved_classes) == 0:
            reserved_classes = None
        else:
            assert len(reserved_classes) == 1
            reserved_classes = reserved_classes[0]

        model_list = model_path.split(',')
        assert len(model_list) == 2
        # first is pose model, second is detection model
        pose_model_path, detection_model_path = model_list

        detection_model_type = model_config['detection'].pop('model_type')
        if detection_model_type == 'TorchYoloXPredictor':
            detection_predictor_config = dict(
                type=detection_model_type,
                model_path=detection_model_path,
                model_config=model_config['detection'])
        else:
            detection_predictor_config = dict(
                model_path=detection_model_path, **model_config['detection'])

        pose_kwargs = model_config['pose']
        pose_kwargs.pop('format', None)

        super().__init__(
            model_path=pose_model_path,
            detection_predictor_config=detection_predictor_config,
            cat_id=reserved_classes,
            **pose_kwargs,
        )

    def get_output_processor(self):
        return _TorchPoseTopDownOutputProcessor()

    def show_result(self,
                    image_path,
                    keypoints,
                    radius=4,
                    thickness=1,
                    kpt_score_thr=0.3,
                    bbox_color='green',
                    show=False,
                    save_path=None):
        dataset_info = self.dataset_info
        # get dataset info
        if (dataset_info is None and hasattr(self.model, 'cfg')
                and 'dataset_info' in self.model.cfg):
            dataset_info = DatasetInfo(self.model.cfg.dataset_info)

        if not dataset_info:
            raise ValueError('Please provide `dataset_info`!')

        skeleton = dataset_info.skeleton
        pose_kpt_color = dataset_info.pose_kpt_color
        pose_link_color = dataset_info.pose_link_color

        if hasattr(self.model, 'module'):
            self.model = self.model.module

        img = self.model.show_result(
            image_path,
            keypoints,
            skeleton,
            radius=radius,
            thickness=thickness,
            pose_kpt_color=pose_kpt_color,
            pose_link_color=pose_link_color,
            kpt_score_thr=kpt_score_thr,
            bbox_color=bbox_color,
            show=show,
            out_file=save_path)

        return img



# Copyright (c) Alibaba, Inc. and its affiliates.
"""
Mostly copy-paste from timm library.
https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/vision_transformer.py

"""
from functools import partial

import torch
import torch.nn as nn
from timm.models.layers import trunc_normal_

from easycv.models.utils import DropPath, Mlp
from ..registry import BACKBONES


def hydra(q, k, v):
    """ Hydra Attention

    Paper link: https://arxiv.org/pdf/2209.07484.pdf (Hydra Attention: Efficient Attention with Many Heads)

    Args:
        q, k, and v should all be tensors of shape
            [batch, tokens, features]
    """
    q = q / q.norm(dim=-1, keepdim=True)
    k = k / k.norm(dim=-1, keepdim=True)
    kv = (k * v).sum(dim=-2, keepdim=True)
    out = q * kv
    return out


class Attention(nn.Module):

    def __init__(self,
                 dim,
                 num_heads=8,
                 qkv_bias=False,
                 qk_scale=None,
                 attn_drop=0.,
                 proj_drop=0.,
                 hydra_attention=False):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.hydra_attention = hydra_attention

    def forward(self, x, rel_pos_bias=None):
        B, N, C = x.shape

        if self.hydra_attention:
            qkv = self.qkv(x).reshape(B, N, 3,
                                      self.num_heads).permute(2, 0, 1, 3)
            q, k, v = qkv[0], qkv[1], qkv[2]

            x = hydra(q, k, v)
            x = x.reshape(B, N, C)

            x = self.proj(x)
            x = self.proj_drop(x)
            return x, None
        else:
            qkv = self.qkv(x).reshape(B, N, 3, self.num_heads,
                                      C // self.num_heads).permute(
                                          2, 0, 3, 1, 4)
            q, k, v = qkv[0], qkv[1], qkv[2]

            attn = (q @ k.transpose(-2, -1)) * self.scale

            if rel_pos_bias is not None:
                attn = attn + rel_pos_bias

            attn = attn.softmax(dim=-1)
            attn = self.attn_drop(attn)

            x = (attn @ v).transpose(1, 2).reshape(B, N, C)

            x = self.proj(x)
            x = self.proj_drop(x)
            return x, attn


class Block(nn.Module):

    def __init__(self,
                 dim,
                 num_heads,
                 mlp_ratio=4.,
                 qkv_bias=False,
                 qk_scale=None,
                 drop=0.,
                 attn_drop=0.,
                 drop_path=0.,
                 act_layer=nn.GELU,
                 norm_layer=nn.LayerNorm,
                 use_layer_scale=False,
                 init_values=1e-4,
                 hydra_attention=False):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
            hydra_attention=hydra_attention)
        self.drop_path = DropPath(
            drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop)
        self.use_layer_scale = use_layer_scale
        if self.use_layer_scale:
            self.gamma_1 = nn.Parameter(
                init_values * torch.ones((dim)), requires_grad=True)
            self.gamma_2 = nn.Parameter(
                init_values * torch.ones((dim)), requires_grad=True)

    def forward(self, x, return_attention=False, rel_pos_bias=None):
        y, attn = self.attn(self.norm1(x), rel_pos_bias=rel_pos_bias)
        if return_attention:
            return attn
        if self.use_layer_scale:
            x = x + self.drop_path(self.gamma_1 * y)
            x = x + self.drop_path(self.gamma_2 * self.mlp(self.norm2(x)))
        else:
            x = x + self.drop_path(y)
            x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x

    def forward_fea_and_attn(self, x):
        y, attn = self.attn(self.norm1(x))
        if self.use_layer_scale:
            x = x + self.drop_path(self.gamma_1 * y)
            x = x + self.drop_path(self.gamma_2 * self.mlp(self.norm2(x)))
        else:
            x = x + self.drop_path(y)
            x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x, attn


class PatchEmbed(nn.Module):
    """ Image to Patch Embedding
    """

    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        num_patches = (img_size // patch_size) * (img_size // patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)

    def forward(self, x):
        B, C, H, W = x.shape
        x = self.proj(x).flatten(2).transpose(1, 2)
        return x


@BACKBONES.register_module
class VisionTransformer(nn.Module):
    """ DeiT III is based on ViT. It uses some strategies to make the vit model
    better, just like layer scale, stochastic depth, 3-Augment.

    Paper link: https://arxiv.org/pdf/2204.07118.pdf (DeiT III: Revenge of the ViT)

    Args:
        img_size (list): Input image size. img_size=[224] means the image size is
            224*224. img_size=[192, 224] means the image size is 192*224.
        patch_size (int): The patch size. Default: 16
        in_chans (int): The num of input channels. Default: 3
        num_classes (int): The num of picture classes. Default: 1000
        embed_dim (int): The dimensions of embedding. Default: 768
        depth (int): The num of blocks. Default: 12
        num_heads (int): Parallel attention heads. Default: 12
        mlp_ratio (float): Mlp expansion ratio. Default: 4.0
        qkv_bias (bool): Does kqv use bias. Default: False
        qk_scale (float | None): In the step of self-attention, if qk_scale is not
            None, it will use qk_scale to scale the q @ k. Otherwise it will use
            head_dim**-0.5 instead of qk_scale. Default: None
        drop_rate (float): Probability of an element to be zeroed after the feed
            forward layer. Default: 0.0
        drop_path_rate (float): Stochastic depth rate. Default: 0
        norm_layer (nn.Module): normalization layer
        global_pool (bool): Global pool before head. Default: False
        use_layer_scale (bool): If use_layer_scale is True, it will use layer
            scale. Default: False
        init_scale (float): It is used for layer scale in Block to scale the
            gamma_1 and gamma_2.
        hydra_attention (bool): If hydra_attention is True, it will use Hydra
            Attention. Default: False
        hydra_attention_layers (int | None): The number of layers that use Hydra Attention.
            If it is None and hydra_attention is True, it will be equal to depth.
            Default: None
        use_dpr_linspace (bool): If use_dpr_linspace is False, all block's drop_path_rate
            are the same. Otherwise, it will use "torch.linspace" on drop_path_rate.
            Default: True

    """

    def __init__(self,
                 img_size=[224],
                 patch_size=16,
                 in_chans=3,
                 num_classes=1000,
                 embed_dim=768,
                 depth=12,
                 num_heads=12,
                 mlp_ratio=4.,
                 qkv_bias=False,
                 qk_scale=None,
                 drop_rate=0.,
                 attn_drop_rate=0.,
                 drop_path_rate=0.,
                 norm_layer=partial(nn.LayerNorm, eps=1e-6),
                 global_pool=False,
                 use_layer_scale=False,
                 init_scale=1e-4,
                 hydra_attention=False,
                 hydra_attention_layers=None,
                 use_dpr_linspace=True,
                 **kwargs):
        super().__init__()

        if hydra_attention:
            if hydra_attention_layers is None:
                hydra_attention_layers = depth
            elif hydra_attention_layers > depth:
                raise ValueError(
                    'When using Hydra Attention, hydra_attention_Layers must be smaller than or equal to depth.'
                )

        self.num_features = self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.mlp_ratio = mlp_ratio
        self.qkv_bias = qkv_bias
        self.qk_scale = qk_scale
        self.drop_rate = drop_rate
        self.attn_drop_rate = attn_drop_rate
        self.norm_layer = norm_layer
        self.use_layer_scale = use_layer_scale
        self.init_scale = init_scale
        self.hydra_attention = hydra_attention
        self.hydra_attention_layers = hydra_attention_layers
        self.drop_path_rate = drop_path_rate
        self.depth = depth

        self.patch_embed = PatchEmbed(
            img_size=img_size[0],
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=embed_dim)
        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches, embed_dim))
        self.pos_drop = nn.Dropout(p=drop_rate)

        if use_dpr_linspace:
            dpr = [
                x.item()
                for x in torch.linspace(0, self.drop_path_rate, self.depth)
            ]
        else:
            dpr = [drop_path_rate for x in range(self.depth)]
        self.dpr = dpr

        if self.hydra_attention:
            hy = [
                x >= (self.depth - self.hydra_attention_layers)
                for x in range(self.depth)
            ]
            head = [
                self.embed_dim if x >=
                (self.depth - self.hydra_attention_layers) else self.num_heads
                for x in range(self.depth)
            ]
        else:
            hy = [False for x in range(self.depth)]
            head = [self.num_heads for x in range(self.depth)]
        self.blocks = nn.ModuleList([
            Block(
                dim=embed_dim,
                num_heads=head[i],
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[i],
                norm_layer=norm_layer,
                use_layer_scale=use_layer_scale,
                init_values=init_scale,
                hydra_attention=hy[i]) for i in range(depth)
        ])
        self.norm = norm_layer(embed_dim)

        # Classifier head
        self.head = nn.Linear(
            embed_dim, num_classes) if num_classes > 0 else nn.Identity()

        # Use global average pooling
        self.global_pool = global_pool
        if self.global_pool:
            self.fc_norm = norm_layer(embed_dim)
            self.norm = None

    def init_weights(self):
        trunc_normal_(self.pos_embed, std=.02)
        trunc_normal_(self.cls_token, std=.02)

        for m in self.modules():
            if isinstance(m, nn.Linear):
                trunc_normal_(m.weight, std=.02)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

    def forward(self, x):

        x = self.forward_features(x)
        x = self.pos_drop(x)
        x = self.head(x)

        return [x]

    def forward_features(self, x):
        B = x.shape[0]
        x = self.patch_embed(x)

        cls_tokens = self.cls_token.expand(B, -1, -1)

        x = x + self.pos_embed
        x = torch.cat((cls_tokens, x), dim=1)

        for blk in self.blocks:
            x = blk(x)
        if self.norm is not None:
            x = self.norm(x)

        if self.global_pool:
            x = x[:, 1:, :].mean(dim=1)
            return self.fc_norm(x)
        else:
            return x[:, 0]



# Copyright (c) OpenMMLab. All rights reserved.

import json
import os
from glob import glob
from multiprocessing import cpu_count

import numpy as np

from easycv.datasets.registry import DATASOURCES
from easycv.file import io
from .base import DetSourceBase


def parse_json(source_item, classes):
    img_path, target_path = source_item
    with io.open(target_path, 'r') as t:
        info = json.load(t)
        img_name = info.get('asset')['name']
        gt_bboxes = []
        gt_labels = []
        for obj in info.get('regions'):
            cls_id = classes.index(obj['tags'][0])
            bbox = obj['boundingBox']

            box = [
                float(bbox['left']),
                float(bbox['top']),
                float(bbox['left'] + bbox['width']),
                float(bbox['top'] + bbox['height'])
            ]
            gt_bboxes.append(box)
            gt_labels.append(cls_id)

    if len(gt_bboxes) == 0:
        gt_bboxes = np.zeros((0, 4), dtype=np.float32)

    img_info = {
        'gt_bboxes': np.array(gt_bboxes, dtype=np.float32),
        'gt_labels': np.array(gt_labels, dtype=np.int64),
        'filename':
        os.path.dirname(target_path).replace('annotations', img_name)
    }

    return img_info


@DATASOURCES.register_module
class DetSourceArtaxor(DetSourceBase):
    """
    data dir is as follows:
    ```
    |- data
        |-Images
            |-000040.jpg
            |-...
        |-Annotations
            |-000040.jpg.txt
            |-...
        |-train.txt
        |-val.txt
        |-...

    ```
    Example1:
        data_source = DetSourceWiderPerson(
            path='/your/data/train.txt',
            classes=${VOC_CLASSES},
        )
    Example1:
        data_source = DetSourceWiderPerson(
            path='/your/voc_data/train.txt',
            classes=${CLASSES},
            img_root_path='/your/data/Images',
            img_root_path='/your/data/Annotations'
        )
    """
    CLASSES = [
        'Araneae', 'Coleoptera', 'Diptera', 'Hemiptera', 'Hymenoptera',
        'Lepidoptera', 'Odonata'
    ]

    def __init__(self,
                 path,
                 classes=CLASSES,
                 cache_at_init=False,
                 cache_on_the_fly=False,
                 label_suffix='.json',
                 parse_fn=parse_json,
                 num_processes=int(cpu_count() / 2),
                 **kwargs) -> None:
        """
        Args:
            path: path of img id list file in root
            classes: classes list
            cache_at_init: if set True, will cache in memory in __init__ for faster training
            cache_on_the_fly: if set True, will cache in memroy during training
            label_suffix: suffix of label file
            parse_fn: parse function to parse item of source iterator
            num_processes: number of processes to parse samples
        """

        self.path = path
        self.label_suffix = label_suffix

        super(DetSourceArtaxor, self).__init__(
            classes=classes,
            cache_at_init=cache_at_init,
            cache_on_the_fly=cache_on_the_fly,
            parse_fn=parse_fn,
            num_processes=num_processes)

    def get_source_iterator(self):

        assert os.path.exists(self.path), f'{self.path} is not exists'

        imgs_path_list = []
        labels_path_list = []

        for category in self.CLASSES:
            img_path = os.path.join(self.path, category)
            assert os.path.exists(img_path), f'{img_path} is not exists'
            label_list = glob(
                os.path.join(img_path, 'annotations', '*' + self.label_suffix))
            for label_path in label_list:

                imgs_path_list.append(category)
                labels_path_list.append(label_path)

        return list(zip(imgs_path_list, labels_path_list))



# Copyright (c) Alibaba, Inc. and its affiliates.
import glob
import math
import os

import numpy as np
import torch
from PIL import Image

from easycv.file import io
from easycv.framework.errors import ValueError
from easycv.utils.checkpoint import load_checkpoint
from easycv.utils.misc import deprecated
from .base import InputProcessor, OutputProcessor, Predictor, PredictorV2
from .builder import PREDICTORS


# onnx specific
def onnx_to_numpy(tensor):
    return tensor.detach().cpu().numpy(
    ) if tensor.requires_grad else tensor.cpu().numpy()


class ClsInputProcessor(InputProcessor):
    """Process inputs for classification models.

    Args:
        cfg (Config): Config instance.
        pipelines (list[dict]): Data pipeline configs.
        batch_size (int): batch size for forward.
        pil_input (bool): Whether use PIL image. If processor need PIL input, set true, default false.
        threads (int): Number of processes to process inputs.
        mode (str): The image mode into the model.
    """

    def __init__(self,
                 cfg,
                 pipelines=None,
                 batch_size=1,
                 pil_input=True,
                 threads=8,
                 mode='BGR'):

        super(ClsInputProcessor, self).__init__(
            cfg, pipelines=pipelines, batch_size=batch_size, threads=threads)
        self.mode = mode
        self.pil_input = pil_input

    def _load_input(self, input):
        """Load image from file or numpy or PIL object.
        Args:
            input: File path or numpy or PIL object.
        Returns:
           {
                'filename': filename,
                'img': img,
                'img_shape': img_shape,
                'img_fields': ['img']
            }
        """
        if self.pil_input:
            results = {}
            if isinstance(input, str):
                img = Image.open(input)
                if img.mode.upper() != self.mode.upper():
                    img = img.convert(self.mode.upper())
                results['filename'] = input
            else:
                if isinstance(input, np.ndarray):
                    input = Image.fromarray(input)
                # assert isinstance(input, ImageFile.ImageFile)
                img = input
                results['filename'] = None
            results['img'] = img
            results['img_shape'] = img.size
            results['ori_shape'] = img.size
            results['img_fields'] = ['img']
            return results

        return super()._load_input(input)


class ClsOutputProcessor(OutputProcessor):
    """Output processor for processing classification model outputs.

    Args:
        topk (int): Return top-k results. Default: 1.
        label_map (dict): Dict of class id to class name.

    """

    def __init__(self, topk=1, label_map={}):
        self.topk = topk
        self.label_map = label_map
        super(ClsOutputProcessor, self).__init__()

    def __call__(self, inputs):
        """Return top-k results."""
        output_prob = inputs['prob'].data.cpu()
        topk_class = torch.topk(output_prob, self.topk).indices.numpy()
        output_prob = output_prob.numpy()
        batch_results = []
        batch_size = output_prob.shape[0]
        for i in range(batch_size):
            result = {'class': np.squeeze(topk_class[i]).tolist()}
            if isinstance(result['class'], int):
                result['class'] = [result['class']]

            if len(self.label_map) > 0:
                result['class_name'] = [
                    self.label_map[i] for i in result['class']
                ]
                result['class_probs'] = {}
                for l_idx, l_name in enumerate(self.label_map):
                    result['class_probs'][l_name] = output_prob[i][l_idx]

            batch_results.append(result)
        return batch_results


@PREDICTORS.register_module()
class ClassificationPredictor(PredictorV2):
    """Predictor for classification.
    Args:
        model_path (str): Path of model path.
        config_file (Optinal[str]): config file path for model and processor to init. Defaults to None.
        batch_size (int): batch size for forward.
        device (str): Support 'cuda' or 'cpu', if is None, detect device automatically.
        save_results (bool): Whether to save predict results.
        save_path (str): File path for saving results, only valid when `save_results` is True.
        pipelines (list[dict]): Data pipeline configs.
        topk (int): Return top-k results. Default: 1.
        pil_input (bool): Whether use PIL image. If processor need PIL input, set true, default false.
        label_map_path (str): File path of saving labels list.
        input_processor_threads (int): Number of processes to process inputs.
        mode (str): The image mode into the model.
    """

    def __init__(self,
                 model_path,
                 config_file=None,
                 batch_size=1,
                 device=None,
                 save_results=False,
                 save_path=None,
                 pipelines=None,
                 topk=1,
                 pil_input=True,
                 label_map_path=None,
                 input_processor_threads=8,
                 mode='BGR',
                 *args,
                 **kwargs):
        self.topk = topk
        self.pil_input = pil_input
        self.label_map_path = label_map_path

        if model_path.endswith('onnx'):
            self.model_type = 'onnx'
            pwd_model = os.path.dirname(model_path)
            raw_model = glob.glob(
                os.path.join(pwd_model, '*.onnx.config.json'))
            if len(raw_model) != 0:
                config_file = raw_model[0]
            else:
                assert len(
                    raw_model
                ) == 0, 'Please have a file with the .onnx.config.json extension in your directory'
        else:
            self.model_type = 'raw'

        if self.pil_input:
            mode = 'RGB'
        super(ClassificationPredictor, self).__init__(
            model_path,
            config_file=config_file,
            batch_size=batch_size,
            device=device,
            save_results=save_results,
            save_path=save_path,
            pipelines=pipelines,
            input_processor_threads=input_processor_threads,
            mode=mode,
            *args,
            **kwargs)

    def get_input_processor(self):
        return ClsInputProcessor(
            self.cfg,
            pipelines=self.pipelines,
            batch_size=self.batch_size,
            threads=self.input_processor_threads,
            pil_input=self.pil_input,
            mode=self.mode)

    def get_output_processor(self):
        # Adapt to torchvision transforms which process PIL inputs.
        if self.label_map_path is None:
            if 'CLASSES' in self.cfg:
                class_list = self.cfg.get('CLASSES', [])
            elif 'class_list' in self.cfg:
                class_list = self.cfg.get('class_list', [])
            else:
                class_list = []
        else:
            with io.open(self.label_map_path, 'r') as f:
                class_list = f.readlines()
        self.label_map = [i.strip() for i in class_list]

        return ClsOutputProcessor(topk=self.topk, label_map=self.label_map)

    def prepare_model(self):
        """Build model from config file by default.
        If the model is not loaded from a configuration file, e.g. torch jit model, you need to reimplement it.
        """
        if self.model_type == 'raw':
            model = self._build_model()
            model.to(self.device)
            model.eval()
            load_checkpoint(model, self.model_path, map_location='cpu')
            return model
        else:
            import onnxruntime
            if onnxruntime.get_device() == 'GPU':
                onnx_model = onnxruntime.InferenceSession(
                    self.model_path, providers=['CUDAExecutionProvider'])
            else:
                onnx_model = onnxruntime.InferenceSession(self.model_path)

            return onnx_model

    def model_forward(self, inputs):
        """Model forward.
        If you need refactor model forward, you need to reimplement it.
        """
        with torch.no_grad():
            if self.model_type == 'raw':
                outputs = self.model(**inputs, mode='test')
            else:
                outputs = self.model.run(None, {
                    self.model.get_inputs()[0].name:
                    onnx_to_numpy(inputs['img'])
                })[0]
                outputs = dict(prob=torch.from_numpy(outputs))
        return outputs


try:
    from easy_vision.python.inference.predictor import PredictorInterface
except:
    from .interface import PredictorInterface


@deprecated(reason='Please use ClassificationPredictor.')
@PREDICTORS.register_module()
class TorchClassifier(PredictorInterface):

    def __init__(self,
                 model_path,
                 model_config=None,
                 topk=1,
                 label_map_path=None):
        """
    init model

    Args:
      model_path: model file path
      model_config: config string for model to init, in json format
    """
        self.predictor = Predictor(model_path)
        if 'class_list' not in self.predictor.cfg and \
            'CLASSES' not in self.predictor.cfg and \
                label_map_path is None:
            raise ValueError(
                "'label_map_path' need to be set, when ckpt doesn't contain key 'class_list' and 'CLASSES'!"
            )

        if label_map_path is None:
            class_list = self.predictor.cfg.get('class_list', [])
            if len(class_list) < 1:
                class_list = self.predictor.cfg.get('CLASSES', [])
            self.label_map = [i.strip() for i in class_list]
        else:
            class_list = open(label_map_path).readlines()
            self.label_map = [i.strip() for i in class_list]
        self.output_name = ['prob', 'class']
        self.topk = topk if topk < len(class_list) else len(class_list)

    def get_output_type(self):
        """
    in this function user should return a type dict, which indicates
    which type of data should the output of predictor be converted to
    * type json, data will be serialized to json str

    * type image, data will be converted to encode image binary and write to oss file,
      whose name is output_dir/${key}/${input_filename}_${idx}.jpg, where input_filename
      is the base filename extracted from url, key corresponds to the key in the dict of output_type,
      if the type of data indexed by key is a list, idx is the index of element in list, otherwhile ${idx} will be empty

    * type video, data will be converted to encode video binary and write to oss file,

    :: return  {
      'image': 'image',
      'feature': 'json'
    }

    indicating that the image data in the output dict will be save to image
    file and feature in output dict will be converted to json

    """
        return {}

    def batch(self, image_tensor_list):
        return torch.stack(image_tensor_list)

    def predict(self, input_data_list, batch_size=-1):
        """
    using session run predict a number of samples using batch_size

    Args:
      input_data_list:  a list of numpy array, each array is a sample to be predicted
      batch_size: batch_size passed by the caller, you can also ignore this param and
        use a fixed number if you do not want to adjust batch_size in runtime
    Return:
      result: a list of dict, each dict is the prediction result of one sample
        eg, {"output1": value1, "output2": value2}, the value type can be
        python int str float, and numpy array
    """
        num_image = len(input_data_list)
        assert len(
            input_data_list) > 0, 'input images should not be an empty list'
        if batch_size > 0:
            num_batches = int(math.ceil(float(num_image) / batch_size))
            image_list = input_data_list
        else:
            num_batches = 1
            batch_size = len(input_data_list)
            image_list = input_data_list

        outputs_list = []
        for batch_idx in range(num_batches):
            batch_image_list = image_list[batch_idx * batch_size:min(
                (batch_idx + 1) * batch_size, len(image_list))]
            image_tensor_list = self.predictor.preprocess(batch_image_list)
            input_data = self.batch(image_tensor_list)
            output_prob = self.predictor.predict_batch(
                input_data, mode='test')['prob'].data.cpu()

            topk_prob = torch.topk(output_prob, self.topk).values.numpy()
            topk_class = torch.topk(output_prob, self.topk).indices.numpy()
            output_prob = output_prob.numpy()

            for idx in range(len(image_tensor_list)):
                single_result = {}
                single_result['class'] = np.squeeze(topk_class[idx]).tolist()
                if isinstance(single_result['class'], int):
                    single_result['class'] = [single_result['class']]
                single_result['class_name'] = [
                    self.label_map[i] for i in single_result['class']
                ]
                single_result['class_probs'] = {}
                for ldx, i in enumerate(self.label_map):
                    single_result['class_probs'][i] = output_prob[idx][ldx]

                outputs_list.append(single_result)

        return outputs_list



# Copyright (c) Alibaba, Inc. and its affiliates.
from .build_loader import build_dataloader
from .sampler import (DistributedGivenIterationSampler,
                      DistributedGroupSampler, DistributedMPSampler,
                      DistributedSampler, GroupSampler, RASampler)

__all__ = [
    'GroupSampler', 'DistributedGroupSampler', 'build_dataloader',
    'DistributedGivenIterationSampler', 'DistributedMPSampler', 'RASampler',
    'DistributedSampler'
]



# Copyright (c) Alibaba, Inc. and its affiliates.

import copy
import ctypes
import itertools
import logging
import os
import timeit
from contextlib import contextmanager

import numpy as np
import pandas as pd
import torch
import torch_blade
import torch_blade.tensorrt
from torch_blade import optimize

from easycv.framework.errors import RuntimeError

os.environ['DISC_ENABLE_STITCH'] = os.environ.get('DISC_ENABLE_STITCH', 'true')
os.environ['DISC_EXPERIMENTAL_SPECULATION_TLP_ENHANCE'] = os.environ.get(
    'DISC_EXPERIMENTAL_SPECULATION_TLP_ENHANCE', 'true')

_cudart = ctypes.CDLL('libcudart.so')


def blade_env_assert():
    env_flag = True

    try:
        import torch
        torch_version = torch.__version__
        torch_cuda = torch.version.cuda
    except:
        torch_version = 'failed'
        torch_cuda = 'failed'
        env_flag = False
        logging.error(
            'import torch and torch cuda failed, please install pytorch with cuda correctly'
        )

    try:
        import torch_blade
    except:
        env_flag = False
        logging.error(
            'Import torch_blade failed, please reference to https://help.aliyun.com/document_detail/205134.html'
        )
        logging.info(
            'Info: your torch version is %s, your torch cuda version is %s' %
            (torch_version, torch_cuda))

    try:
        import torch_blade.tensorrt
    except:
        env_flag = False
        logging.error(
            'Import torch_blade.tensorrt failed, Install torch_blade.tensorrt and export  xx/tensorrt.so to your python ENV'
        )

    logging.info(
        'Welcome to use torch_blade, with torch %s, cuda %s, blade %s' %
        (torch_version, torch_cuda, torch_blade.version.__version__))

    return env_flag


@contextmanager
def opt_trt_config(
        input_config=dict(enable_fp16=True, fp16_fallback_op_ratio=0.05)):
    from torch_blade import tensorrt
    torch_config = torch_blade.Config()

    BLADE_CONFIG_DEFAULT = dict(
        optimization_pipeline='TensorRT',
        enable_fp16=True,
        customize_op_black_list=[
            # 'aten::select', 'aten::index', 'aten::slice', 'aten::view', 'aten::upsample'
        ],
        fp16_fallback_op_ratio=0.05,
        preserved_attributes=[],
    )
    BLADE_CONFIG_KEYS = list(BLADE_CONFIG_DEFAULT.keys())

    for key in BLADE_CONFIG_DEFAULT.keys():
        setattr(torch_config, key, BLADE_CONFIG_DEFAULT[key])
        logging.info('setting blade torch_config %s to %s by default' %
                     (key, BLADE_CONFIG_DEFAULT[key]))

    for key in input_config.keys():
        if key in BLADE_CONFIG_KEYS:
            setattr(torch_config, key, input_config[key])
            logging.warning(
                'setting blade torch_config %s to %s by user config' %
                (key, input_config[key]))

    try:
        with torch_config:
            yield
    finally:
        pass


def cu_prof_start():
    ret = _cudart.cudaProfilerStart()
    if ret != 0:
        raise RuntimeError('cudaProfilerStart() returned %d' % ret)


def cu_prof_stop():
    ret = _cudart.cudaProfilerStop()
    if ret != 0:
        raise RuntimeError('cudaProfilerStop() returned %d' % ret)


@contextmanager
def opt_blade_mixprec():
    try:
        dummy = torch.cuda.amp.autocast(True)
        yield
    finally:
        pass


@contextmanager
def opt_disc_config(enable_fp16=True):
    torch_config = torch_blade.config.Config()
    torch_config.enable_fp16 = enable_fp16
    try:
        with torch_config:
            yield
    finally:
        pass


def computeStats(backend, timings, batch_size=1, model_name='default'):
    """
    compute the statistical metric of time and speed

    Args:
        backend (str):  backend name
        timings (List): time list
        batch_size (int)： image batch
        model_name (str): tested model name
    """
    times = np.array(timings)
    steps = len(times)
    speeds = batch_size / times
    time_mean = np.mean(times)
    time_med = np.median(times)
    time_99th = np.percentile(times, 99)
    time_std = np.std(times, ddof=0)
    speed_mean = np.mean(speeds)
    speed_med = np.median(speeds)

    msg = ('\n%s =================================\n'
           'batch size=%d, num iterations=%d\n'
           '  Median FPS: %.1f, mean: %.1f\n'
           '  Median latency: %.6f, mean: %.6f, 99th_p: %.6f, std_dev: %.6f\n'
           ) % (
               backend,
               batch_size,
               steps,
               speed_med,
               speed_mean,
               time_med,
               time_mean,
               time_99th,
               time_std,
           )

    meas = {
        'Name': model_name,
        'Backend': backend,
        'Median(FPS)': speed_med,
        'Mean(FPS)': speed_mean,
        'Median(ms)': time_med,
        'Mean(ms)': time_mean,
        '99th_p': time_99th,
        'std_dev': time_std,
    }

    return meas


@torch.no_grad()
def benchmark(model,
              inputs,
              backend,
              batch_size,
              model_name='default',
              num_iters=200,
              warmup_iters=5,
              fp16=False):
    """
    evaluate the time and speed of different models

    Args:
        model: input model
        inputs: input of the model
        backend (str):  backend name
        batch_size (int)： image batch
        model_name (str): tested model name
        num_iters: test forward times
    """
    for _ in range(warmup_iters):
        if fp16:
            with torch.cuda.amp.autocast():
                model(*copy.deepcopy(inputs))
        else:
            model(*copy.deepcopy(inputs))

    torch.cuda.synchronize()
    timings = []
    for i in range(num_iters):
        start_time = timeit.default_timer()
        if fp16:
            with torch.cuda.amp.autocast():
                model(*copy.deepcopy(inputs))
        else:
            model(*copy.deepcopy(inputs))
        torch.cuda.synchronize()
        end_time = timeit.default_timer()
        meas_time = end_time - start_time
        timings.append(meas_time)

    return computeStats(backend, timings, batch_size, model_name)


def collect_tensors(data):
    if isinstance(data, torch.Tensor):
        return [data]
    elif isinstance(data, list):
        return list(itertools.chain(*[collect_tensors(d) for d in data]))
    elif isinstance(data, dict):
        sorted_pairs = sorted(data.items(), key=lambda x: x[0])
        sorted_list = [v for k, v in sorted_pairs]
        return collect_tensors(sorted_list)
    elif isinstance(data, tuple):
        return collect_tensors(list(data))
    else:
        return []


def check_results(results0, results1):
    from torch_blade.testing.common_utils import assert_almost_equal

    results0 = collect_tensors(results0)
    results1 = collect_tensors(results1)

    try:
        assert_almost_equal(results0, results1, rtol=1e-3, atol=1e-3)
        logging.info('Accuraccy check passed')
    except Exception as err:
        logging.error(err)


def blade_optimize(speed_test_model,
                   model,
                   inputs,
                   blade_config=dict(
                       enable_fp16=True, fp16_fallback_op_ratio=0.05),
                   backend='TensorRT',
                   batch=1,
                   warmup_iters=10,
                   compute_cost=True,
                   use_profile=False,
                   check_result=False,
                   static_opt=True,
                   min_num_nodes=None,
                   check_inputs=True,
                   fp16=False):

    if not static_opt:
        logging.info(
            'PAI-Blade use dynamic optimize for input model, export model is build for dynamic shape input'
        )
        optimize_op = optimize
    else:
        logging.info(
            'PAI-Blade use static optimize for input model, export model must be used as static shape input'
        )
        from torch_blade.optimization import _static_optimize
        optimize_op = _static_optimize
    if min_num_nodes is not None:
        import torch_blade.clustering.support_fusion_group as blade_fusion
        with blade_fusion.min_group_nodes(min_num_nodes=min_num_nodes):
            with opt_trt_config(blade_config):
                opt_model = optimize_op(
                    model,
                    allow_tracing=True,
                    model_inputs=tuple(copy.deepcopy(inputs)),
                )
    else:
        with opt_trt_config(blade_config):
            opt_model = optimize_op(
                model,
                allow_tracing=True,
                model_inputs=tuple(copy.deepcopy(inputs)),
            )
    if compute_cost:
        logging.info('Running benchmark...')
        results = []
        inputs_t = copy.deepcopy(inputs)

        # end2end model and scripts needs different channel purmulate, encounter this problem only when we use end2end export
        if check_inputs and (inputs_t[0].shape[-1] == 3):
            shape_length = len(inputs_t[0].shape)
            if shape_length == 4:
                inputs_t = inputs_t[0].permute(0, 3, 1, 2)
                inputs_t = [inputs_t]

            if shape_length == 3:
                inputs_t = inputs_t[0].permute(2, 0, 1)
                inputs_t = (torch.unsqueeze(inputs_t, 0), )

        results.append(
            benchmark(
                speed_test_model,
                inputs_t,
                backend,
                batch,
                'easycv',
                warmup_iters=warmup_iters,
                fp16=fp16))
        results.append(
            benchmark(
                model,
                copy.deepcopy(inputs),
                backend,
                batch,
                'easycv script',
                warmup_iters=warmup_iters,
                fp16=fp16))
        results.append(
            benchmark(
                opt_model,
                copy.deepcopy(inputs),
                backend,
                batch,
                'blade',
                warmup_iters=warmup_iters,
                fp16=fp16))

        logging.info('Model Summary:')
        summary = pd.DataFrame(results)
        print(summary.to_markdown())

    if use_profile:
        torch.cuda.empty_cache()
        # warm-up
        for k in range(warmup_iters):
            test_result = opt_model(*copy.deepcopy(inputs))
            torch.cuda.synchronize()

        torch.cuda.synchronize()
        cu_prof_start()
        for k in range(warmup_iters):
            test_result = opt_model(*copy.deepcopy(inputs))
            torch.cuda.synchronize()
        cu_prof_stop()
        import torch.autograd.profiler as profiler
        with profiler.profile(use_cuda=True) as prof:
            for k in range(warmup_iters):
                test_result = opt_model(*copy.deepcopy(inputs))
                torch.cuda.synchronize()

        with profiler.profile(use_cuda=True) as prof:
            for k in range(warmup_iters):
                test_result = opt_model(*copy.deepcopy(inputs))
                torch.cuda.synchronize()

        prof_str = prof.key_averages().table(sort_by='cuda_time_total')
        print(f'{prof_str}')

    if check_result:
        output = model(*copy.deepcopy(inputs))
        test_result = opt_model(*copy.deepcopy(inputs))
        check_results(output, test_result)

    return opt_model



# Copyright (c) Alibaba, Inc. and its affiliates.
import fnmatch
import logging
import os
import re
import time
from datetime import datetime, timedelta
from io import BytesIO, StringIO
from typing import List, Union

from tqdm import tqdm
from tqdm.utils import CallbackIOWrapper

from easycv.framework.errors import (FileNotFoundError, IOError, RuntimeError,
                                     ValueError)
from .base import IOLocal
from .utils import (OSS_PREFIX, create_namedtuple, get_oss_config, is_oss_path,
                    mute_stderr, oss_progress)


def set_oss_env(ak_id: str, ak_secret: str, hosts: Union[str, List[str]],
                buckets: Union[str, List[str]]):
    if isinstance(buckets, str):
        buckets = [buckets]
    if isinstance(hosts, str):
        hosts = [hosts]
    assert (len(buckets) == len(hosts))

    os.environ['OSS_ACCESS_KEY_ID'] = ak_id
    os.environ['OSS_ACCESS_KEY_SECRET'] = ak_secret
    os.environ['OSS_ENDPOINTS'] = ','.join(hosts)
    os.environ['OSS_BUCKETS'] = ','.join(buckets)


class IO(IOLocal):
    """
    IO module to support both local and oss io.
    If access oss file, you need to authorize OSS, please refer to `IO.access_oss`.
    """

    __name__ = 'IO'

    def __init__(self, max_retry=10, retry_wait=0.1, max_retry_wait=30):

        super(IO, self).__init__()

        self.oss_pattern = re.compile(r'oss://([^/]+)/(.+)')
        self.buckets_map = {}
        self.auth = None
        self.oss_config = None
        self.max_retry = max_retry
        self.retry_wait = retry_wait
        self.max_retry_wait = max_retry_wait

        if self._is_oss_env_prepared():
            self.access_oss(
                ak_id=os.environ.get('OSS_ACCESS_KEY_ID'),
                ak_secret=os.environ.get('OSS_ACCESS_KEY_SECRET'),
                hosts=[
                    i.strip()
                    for i in os.environ.get('OSS_ENDPOINTS').split(',')
                ],
                buckets=[
                    i.strip() for i in os.environ.get('OSS_BUCKETS').split(',')
                ])

    def _is_oss_env_prepared(self):
        return (os.environ.get('OSS_ACCESS_KEY_ID')
                and os.environ.get('OSS_ACCESS_KEY_SECRET')
                and os.environ.get('OSS_ENDPOINTS')
                and os.environ.get('OSS_BUCKETS'))

    def _get_default_oss_config(self):
        oss_config = get_oss_config()
        ak_id = oss_config['ak_id']
        ak_secret = oss_config['ak_secret']
        hosts = oss_config['hosts']
        buckets = oss_config['buckets']
        return ak_id, ak_secret, hosts, buckets

    def access_oss(self,
                   ak_id: str = '',
                   ak_secret: str = '',
                   hosts: Union[str, List[str]] = '',
                   buckets: Union[str, List[str]] = ''):
        """If access oss file, you need to authorize OSS as follows:

        Method1:
        from easycv.file import io
        io.access_oss(
            ak_id='your_accesskey_id',
            ak_secret='your_accesskey_secret',
            hosts='your endpoint' or ['your endpoint1', 'your endpoint2'],
            buckets='your bucket' or ['your bucket1', 'your bucket2'])
        Method2:
            Add oss config to your local file `~/.ossutilconfig`, as follows:
            More oss config information, please refer to: https://help.aliyun.com/document_detail/120072.html
            ```
            [Credentials]
                language = CH
                endpoint = your endpoint
                accessKeyID = your_accesskey_id
                accessKeySecret = your_accesskey_secret
            [Bucket-Endpoint]
                bucket1 = endpoint1
                bucket2 = endpoint2
            ```
            Then run the following command, the config file will be read by default to authorize oss.

            from easycv.file import io
            io.access_oss()
        """
        from oss2 import Auth
        if not (ak_id and ak_secret):
            try:
                ak_id, ak_secret, hosts, buckets = \
                    self._get_default_oss_config()
            except Exception as e:
                logging.error(e)
                raise ValueError(
                    'Please provide oss config or create `~/.ossutilconfig` file!'
                )
        self.auth = Auth(ak_id, ak_secret)
        if isinstance(buckets, str):
            buckets = [buckets]
        if isinstance(hosts, str):
            hosts = [hosts for _ in range(len(buckets))]
        else:
            assert len(hosts) == len(buckets), \
                'number of hosts and number of buckets should be the same'

        self.buckets_map = {
            bucket_name: self._create_bucket_obj(host, bucket_name)
            for host, bucket_name in zip(hosts, buckets)
        }

        self.oss_config = create_namedtuple(
            ak_id=ak_id, ak_secret=ak_secret, hosts=hosts, buckets=buckets)

        if not self._is_oss_env_prepared():
            set_oss_env(
                ak_id=ak_id, ak_secret=ak_secret, hosts=hosts, buckets=buckets)

    def _create_bucket_obj(self, host, bucket_name):
        import oss2
        from oss2 import Bucket

        try:
            return Bucket(self.auth, host, bucket_name)
        except oss2.exceptions.ClientError as e:
            logging.error(f'Invalid bucket name "{bucket_name}"')
            raise oss2.exceptions.ClientError(e)

    def _get_bucket_obj_and_path(self, path):
        m = self.oss_pattern.match(path)
        if not m:
            raise IOError(
                f'invalid oss path: "{path}", should be "oss://<bucket_name>/path"'
            )
        bucket_name, path = m.groups()
        path = path.replace('//', '/')
        bucket_name = bucket_name.split('.')[0]

        bucket = self.buckets_map.get(bucket_name, None)

        if not bucket:
            raise IOError(
                f'Bucket {bucket_name} not registered in oss_io_config')
        return bucket, path

    def _read_oss(self, full_path, mode):
        assert mode in ['r', 'rb']
        path_exists = self.exists(full_path)
        if not path_exists:
            raise FileNotFoundError(full_path)
        bucket, path = self._get_bucket_obj_and_path(full_path)
        num_retry = 0
        retry_wait = self.retry_wait

        data = None
        while num_retry < self.max_retry:
            try:
                obj = bucket.get_object(path)
                # auto cache large files to avoid memory issues
                if obj.content_length > 200 * 1024**2:  # 200M
                    with tqdm(
                            total=obj.content_length,
                            unit='B',
                            unit_scale=True,
                            unit_divisor=1024,
                            leave=False,
                            desc='reading ' +
                            os.path.basename(full_path)) as t:
                        obj = CallbackIOWrapper(t.update, obj, 'read')
                        data = obj.read()
                else:
                    data = obj.read()
                break
            except Exception as e:
                num_retry += 1
                logging.warning(
                    f'read exception occur, sleep {retry_wait}s to retry num_retry/max_retry {num_retry}/{self.max_retry}\n {e}'
                )
                if retry_wait > 0:
                    time.sleep(retry_wait)
                    retry_wait = min(retry_wait * 2, self.max_retry_wait)

        if data is None:
            raise IOError('Read file error: %s!' % full_path)

        if mode == 'rb':
            return NullContextWrapper(BytesIO(data))
        else:
            return NullContextWrapper(StringIO(data.decode()))

    def _write_oss(self, full_path, mode):
        assert mode in ['w', 'wb']
        bucket, path = self._get_bucket_obj_and_path(full_path)
        path_exists = self.exists(full_path)
        if path_exists:
            bucket.delete_object(path)
        if mode == 'wb':
            return BinaryOSSFile(bucket, path)
        return OSSFile(bucket, path)

    def _append_oss(self, full_path):
        path_exists = self.exists(full_path)
        bucket, path = self._get_bucket_obj_and_path(full_path)
        position = bucket.head_object(path).content_length \
            if path_exists else 0
        return OSSFile(bucket, path, position=position)

    def open(self, full_path, mode='r'):
        """
        Same usage as the python build-in `open`.
        Support local path and oss path.

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            # Write something to a oss file.
            with io.open('oss://bucket_name/demo.txt', 'w') as f:
                f.write("test")

            # Read from a oss file.
            with io.open('oss://bucket_name/demo.txt', 'r') as f:
                print(f.read())
        Args:
            full_path: absolute oss path
        """
        if not is_oss_path(full_path):
            return super().open(full_path, mode)

        if 'w' in mode:
            return self._write_oss(full_path, mode)
        elif mode == 'a':
            return self._append_oss(full_path)
        elif 'r' in mode:
            return self._read_oss(full_path, mode)
        else:
            raise ValueError('invalid mode: %s' % mode)

    def exists(self, path):
        """
        Whether the file exists, same usage as `os.path.exists`.
        Support local path and oss path.

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            ret = io.exists('oss://bucket_name/dir')
            print(ret)
        Args:
            path: oss path or local path
        """
        if not is_oss_path(path):
            return super().exists(path)

        bucket, _path = self._get_bucket_obj_and_path(path)
        # TODO: 网络不稳时try，其他错误直接抛出
        if not path.endswith('/'):
            # if file exists
            exists = self._obj_exists(bucket, _path)
            # if bucket exists
            if not exists:
                _path = self._correct_oss_dir(_path)
                exists = self._obj_exists(bucket, _path)
        else:
            # if bucket exists
            exists = self._obj_exists(bucket, _path)

        return exists

    def _obj_exists(self, bucket, path):
        return bucket.object_exists(path)

    def move(self, src, dst):
        """Move src to dst,  same usage as shutil.move.
        Support local path and oss path.

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            # move oss file to local
            io.move('oss://bucket_name/file.txt', '/your/local/path/file.txt')
            # move oss file to oss
            io.move('oss://bucket_name/file.txt', 'oss://bucket_name/file.txt')
            # move local file to oss
            io.move('/your/local/file.txt', 'oss://bucket_name/file.txt')
            # move directory
            io.move('oss://bucket_name/dir1', 'oss://bucket_name/dir2')
        Args:
            src: oss path or local path
            dst: oss path or local path
        """
        if not is_oss_path(src) and not is_oss_path(dst):
            return super().move(src, dst)
        if src == dst:
            return
        if self.isfile(src):
            self.copy(src, dst)
        else:
            self.copytree(src, dst)
        self.remove(src)

    def safe_copy(self, src, dst, try_max=3):
        """ oss always bug, we need safe_copy!
        """
        try_flag = True
        try_idx = 0
        while try_flag and try_idx < try_max:
            try_idx += 1
            try:
                self.copy(src, dst)
                try_flag = False
            except:
                pass

        if try_flag:
            print('oss copy from %s to %s failed!!! TRIGGER safe exit!' %
                  (src, dst))
        return

    def _download_oss(self, src, dst):
        target_dir, _ = os.path.split(dst)
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        bucket, src = self._get_bucket_obj_and_path(src)
        obj = bucket.get_object(src)
        if obj.content_length > 100 * 1024**2:  # 100M
            with oss_progress('downloading') as callback:
                bucket.get_object_to_file(src, dst, progress_callback=callback)
        else:
            bucket.get_object_to_file(src, dst)

        return

    def _upload_oss(self, src, dst):
        bucket, dst = self._get_bucket_obj_and_path(dst)
        self._create_oss_dirs(bucket, os.path.split(dst)[0])
        src_size = os.stat(src).st_size
        if src_size > 5 * 1024**3:  # 5G
            raise RuntimeError(
                f'A file > 5G cannot be uploaded to OSS. Please split your file first.\n{src}'
            )
        if src_size > 100 * 1024**2:  # 100M
            with oss_progress('uploading') as callback:
                logging.info(f'upload {src} to {dst}')
                bucket.put_object_from_file(
                    dst, src, progress_callback=callback)
        else:
            bucket.put_object_from_file(dst, src)

        return

    def _copy_oss(self, src, dst):
        bucket, dst = self._get_bucket_obj_and_path(dst)
        # copy between oss paths
        src_bucket, src = self._get_bucket_obj_and_path(src)
        total_size = src_bucket.head_object(src).content_length
        if src_bucket.get_bucket_location(
        ).location != bucket.get_bucket_location().location:
            import tempfile
            local_tmp = os.path.join(tempfile.gettempdir(), src)
            self.copy(f'{OSS_PREFIX}{src_bucket.bucket_name}/{src}', local_tmp)
            self.copy(local_tmp, f'{OSS_PREFIX}{bucket.bucket_name}/{dst}')
            self.remove(local_tmp)
            return

        if total_size < 1024**3 or src_bucket != bucket:  # 1GB
            bucket.copy_object(src_bucket.bucket_name, src, dst)
        else:
            # multipart copy
            from oss2.models import PartInfo
            from oss2 import determine_part_size
            part_size = determine_part_size(
                total_size, preferred_size=100 * 1024)
            upload_id = bucket.init_multipart_upload(dst).upload_id
            parts = []

            part_number = 1
            offset = 0
            while offset < total_size:
                num_to_upload = min(part_size, total_size - offset)
                byte_range = (offset, offset + num_to_upload - 1)

                result = bucket.upload_part_copy(bucket.bucket_name, src,
                                                 byte_range, dst, upload_id,
                                                 part_number)
                parts.append(PartInfo(part_number, result.etag))

                offset += num_to_upload
                part_number += 1

            bucket.complete_multipart_upload(dst, upload_id, parts)

    def copy(self, src, dst):
        """
        Copy a file from src to dst. Same usage as `shutil.copyfile`.
        If you want to copy a directory, please use `easycv.io.copytree`

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            # Copy a file from local to oss:
            io.copy('/your/local/file.txt', 'oss://bucket/dir/file.txt')

            # Copy a oss file to local:
            io.copy('oss://bucket/dir/file.txt', '/your/local/file.txt')

            # Copy a file from oss to oss::
            io.copy('oss://bucket/dir/file.txt', 'oss://bucket/dir/file2.txt')

        Args:
            src: oss path or local path
            dst: oss path or local path
        """
        cloud_src = is_oss_path(src)
        cloud_dst = is_oss_path(dst)
        if not cloud_src and not cloud_dst:
            return super().copy(src, dst)

        if src == dst:
            return
        # download
        if cloud_src and not cloud_dst:
            self._download_oss(src, dst)
            return
        # upload
        if cloud_dst and not cloud_src:
            self._upload_oss(src, dst)
            return

        self._copy_oss(src, dst)

    def copytree(self, src, dst):
        """
        Copy files recursively from src to dst. Same usage as `shutil.copytree`.
        If you want to copy a file, please use `easycv.io.copy`.

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            # copy files from local to oss
            io.copytree(src='/your/local/dir1', dst='oss://bucket_name/dir2')
            # copy files from oss to local
            io.copytree(src='oss://bucket_name/dir2', dst='/your/local/dir1')
            # copy files from oss to oss
            io.copytree(src='oss://bucket_name/dir1', dst='oss://bucket_name/dir2')

        Args:
            src: oss path or local path
            dst: oss path or local path
        """
        cloud_src = is_oss_path(src)
        cloud_dst = is_oss_path(dst)
        if not cloud_src and not cloud_dst:
            return super().copytree(src, dst)

        self.makedirs(dst)
        created_dir = {dst}
        src_files = self.listdir(src, recursive=True)
        max_len = min(max(map(len, src_files)), 50)
        with tqdm(src_files, desc='copytree', leave=False) as progress:
            for file in progress:
                src_file = os.path.join(src, file)
                dst_file = os.path.join(dst, file)
                dst_dir = os.path.dirname(dst_file)
                if dst_dir not in created_dir:
                    self.makedirs(dst_dir)
                    created_dir.add(dst_dir)
                progress.set_postfix({'file': f'{file:-<{max_len}}'[:max_len]})
                self.copy(src_file, dst_file)

    def listdir(self,
                path,
                recursive=False,
                full_path=False,
                contains: Union[str, List[str]] = None):
        """
        List all objects in path. Same usage as `os.listdir`.

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            ret = io.listdir('oss://bucket/dir', recursive=True)
            print(ret)
        Args:
            path: local file path or oss path.
            recursive: If False, only list the top level objects.
                If True, recursively list all objects.
            full_path: if full path, return files with path prefix.
            contains: substr to filter list files.

        return: A list of path.
        """
        from oss2 import ObjectIterator

        if not is_oss_path(path):
            return super().listdir(path, recursive, full_path, contains)

        bucket, path = self._get_bucket_obj_and_path(path)
        path = self._correct_oss_dir(path)
        files = [
            obj.key for obj in ObjectIterator(
                bucket, prefix=path, delimiter='' if recursive else '/')
        ]
        if path in files:
            files.remove(path)
        if recursive:
            files = [
                i for i in files
                if not self.isdir(f'{OSS_PREFIX}{bucket.bucket_name}/{i}')
            ]

        if not files and not self._obj_exists(bucket, path):
            raise FileNotFoundError(
                f'No such directory: {OSS_PREFIX}{bucket.bucket_name}/{path}')

        if full_path:
            files = [
                f'{OSS_PREFIX}{bucket.bucket_name}/{file}' for file in files
            ]
        else:
            files = [file[len(path):] for file in files]

        if not contains:
            return files

        if isinstance(contains, str):
            contains = [contains]
        files = [
            file for file in files
            if any(keyword in file for keyword in contains)
        ]
        return files

    def _remove_obj(self, path):
        bucket, path = self._get_bucket_obj_and_path(path)
        with mute_stderr():
            bucket.delete_object(path)

    def remove(self, path):
        """
        Remove a file or a directory recursively.
        Same usage as `os.remove` or `shutil.rmtree`.

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            # Remove a oss file
            io.remove('oss://bucket_name/file.txt')

            # Remove a oss directory
            io.remove('oss://bucket_name/dir/')

        Args:
            path: local or oss path, file or directory
        """
        if not is_oss_path(path):
            return super().remove(path)

        if self.isfile(path):
            self._remove_obj(path)
        else:
            return self.rmtree(path)

    def _correct_oss_dir(self, path):
        if not path.endswith('/'):
            path += '/'

        return path

    def rmtree(self, path):
        """
        Remove directory recursively, same usage as `shutil.rmtree`

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            io.remove('oss://bucket_name/dir_name')
            # Or
            io.remove('oss://bucket_name/dir_name/')
        Args:
            path: oss path
        """
        if not is_oss_path(path):
            return super().rmtree(path)
        # have to delete its content first before delete the directory itself
        for file in self.listdir(path, recursive=True, full_path=True):
            logging.info(f'delete {file}')
            self._remove_obj(file)
        if self.exists(path):
            # remove the directory itself
            path = self._correct_oss_dir(path)
            self._remove_obj(path)

    def makedirs(self, path, exist_ok=True):
        """
        Create directories recursively, same usage as `os.makedirs`

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            io.makedirs('oss://bucket/new_dir/')
        Args:
            path: local or oss dir path
        """
        del exist_ok
        # there is no need to create directory in oss
        if not is_oss_path(path):
            return super().makedirs(path)
        else:
            bucket, _path = self._get_bucket_obj_and_path(path)
            self._create_oss_dirs(bucket, _path)

    def _create_oss_dirs(self, bucket, dir_path):
        """
        Fix cannot access middle folder created by `bucket.put_object_from_file`
        by put null object to middle dir to create dir object.
        """
        dirs = dir_path.split(os.sep)
        cur_d = ''
        for d in dirs:
            cur_d = os.path.join(cur_d, d)
            cur_d = self._correct_oss_dir(cur_d)
            if not self._obj_exists(bucket, cur_d):
                bucket.put_object(cur_d, '')

    def isdir(self, path):
        """
        Return whether a path is directory, same usage as `os.path.isdir`

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            io.isdir('oss://bucket/dir/')
        Args:
            path: local or oss path
        Return: bool, True or False.
        """
        if not is_oss_path(path):
            return super().isdir(path)
        path = self._correct_oss_dir(path)
        return self.exists(path)

    def isfile(self, path):
        """
        Return whether a path is file object, same usage as `os.path.isfile`

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            io.isfile('oss://bucket/file.txt')
        Args:
            path: local or oss path
        Return: bool, True or False.
        """
        if not is_oss_path(path):
            return super().exists(path) and not super().isdir(path)
        return self.exists(path) and not self.isdir(path)

    def glob(self, file_path):
        """
        Return a list of paths matching a pathname pattern.
        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            io.glob('oss://bucket/dir/*.txt')
        Args:
            path: local or oss file pattern
        Return: list, a list of paths.
        """
        if not is_oss_path(file_path):
            return super().glob(file_path)

        dir_name = self._correct_oss_dir(os.path.split(file_path)[0])

        file_list = self.listdir(dir_name)
        reture_list = []
        for l in file_list:
            name = dir_name + l
            if fnmatch.fnmatch(name, file_path):
                reture_list.append(name)
        return reture_list

    def abspath(self, path):
        if not is_oss_path(path):
            return super().abspath(path)
        return path

    def authorize(self, path):
        if not is_oss_path(path):
            raise ValueError('Only oss path can use "authorize"')
        import oss2
        bucket, path = self._get_bucket_obj_and_path(path)
        bucket.put_object_acl(path, oss2.OBJECT_ACL_PUBLIC_READ)

    def last_modified(self, path):
        if not is_oss_path(path):
            return super().last_modified(path)
        return datetime.strptime(
            self.last_modified_str(path),
            r'%a, %d %b %Y %H:%M:%S %Z') + timedelta(hours=8)

    def last_modified_str(self, path):
        if not is_oss_path(path):
            return super().last_modified_str(path)
        bucket, path = self._get_bucket_obj_and_path(path)
        return bucket.get_object_meta(path).headers['Last-Modified']

    def size(self, path: str) -> int:
        """
        Get the size of file path, same usage as `os.path.getsize`

        Example:
            from easycv.file import io
            io.access_oss(your oss config) # only oss file need, refer to `IO.access_oss`
            size = io.size('oss://bucket/file.txt')
            print(size)
        Args:
            path: local or oss path.
        Return: size of file in bytes
        """
        if not is_oss_path(path):
            return super().size(path)
        bucket, path = self._get_bucket_obj_and_path(path)
        return int(bucket.get_object_meta(path).headers['Content-Length'])


class OSSFile:

    def __init__(self, bucket, path, position=0):
        self.position = position
        self.bucket = bucket
        self.path = path
        self.buffer = StringIO()

    def write(self, content):
        # without a "with" statement, the content is written immediately without buffer
        # when writing a large batch of contents at a time, this will be quite slow
        import oss2
        buffer = self.buffer.getvalue()
        if buffer:
            content = buffer + content
            self.buffer.close()
            self.buffer = StringIO()
        try:
            result = self.bucket.append_object(self.path, self.position,
                                               content)
            self.position = result.next_position
        except oss2.exceptions.PositionNotEqualToLength:
            raise RuntimeError(
                f'Race condition detected. It usually means multiple programs were writing to the same file'
                f'{OSS_PREFIX}{self.bucket.bucket_name}/{self.path} (Error 409: PositionNotEqualToLength)'
            )
        except (oss2.exceptions.RequestError,
                oss2.exceptions.ServerError) as e:
            self.buffer.write(content)
            logging.error(
                str(e) +
                f'when writing to {OSS_PREFIX}{self.bucket.bucket_name}/{self.path}. Content buffered.'
            )

    def flush(self, retry=0):
        import oss2
        try:
            self.bucket.append_object(self.path, self.position,
                                      self.buffer.getvalue())
        except oss2.exceptions.RequestError as e:
            if 'timeout' not in str(e) or retry > 2:
                raise
            # retry if timeout
            logging.error('| OSSIO timeout. Retry uploading...')
            import time
            time.sleep(5)
            self.flush(retry + 1)
        except oss2.exceptions.ObjectNotAppendable as e:
            from . import io
            logging.error(str(e) + '\nTrying to recover..\n')
            full_path = f'{OSS_PREFIX}{self.bucket.bucket_name}/{self.path}'
            with io.open(full_path) as f:
                prev_content = f.read()
            io.remove(full_path)
            self.position = 0
            content = self.buffer.getvalue()
            self.buffer.close()
            self.buffer = StringIO()
            self.write(prev_content)
            self.write(content)

    def close(self):
        self.flush()

    def seek(self, position):
        self.position = position

    def __enter__(self):
        return self.buffer

    def __exit__(self, *args):
        self.close()


class BinaryOSSFile:

    def __init__(self, bucket, path):
        self.bucket = bucket
        self.path = path
        self.buffer = BytesIO()

    def __enter__(self):
        return self.buffer

    def __exit__(self, *args):
        value = self.buffer.getvalue()
        if len(value) > 100 * 1024**2:  # 100M
            with oss_progress('uploading') as callback:
                self.bucket.put_object(
                    self.path, value, progress_callback=callback)
        else:
            self.bucket.put_object(self.path, value)


class NullContextWrapper:

    def __init__(self, obj):
        self._obj = obj

    def __getattr__(self, name):
        return getattr(self._obj, name)

    def __iter__(self):
        return self._obj.__iter__()

    def __next__(self):
        return self._obj.__next__()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass



# Copyright (c) Alibaba, Inc. and its affiliates.
import torch
import torch.nn as nn

from easycv.framework.errors import KeyError
from easycv.utils.checkpoint import load_checkpoint
from easycv.utils.logger import get_root_logger
from .. import builder
from ..base import BaseModel
from ..registry import MODELS


@MODELS.register_module
class BYOL(BaseModel):
    '''BYOL unofficial implementation. Paper: https://arxiv.org/abs/2006.07733
    '''

    def __init__(self,
                 backbone,
                 neck=None,
                 head=None,
                 pretrained=None,
                 base_momentum=0.996,
                 **kwargs):
        super(BYOL, self).__init__()

        self.pretrained = pretrained
        self.online_net = nn.Sequential(
            builder.build_backbone(backbone), builder.build_neck(neck))
        self.target_net = nn.Sequential(
            builder.build_backbone(backbone), builder.build_neck(neck))
        self.backbone = self.online_net[0]
        for param in self.target_net.parameters():
            param.requires_grad = False
        self.head = builder.build_head(head)
        self.init_weights()

        self.base_momentum = base_momentum
        self.momentum = base_momentum

    def init_weights(self):
        if isinstance(self.pretrained, str):
            logger = get_root_logger()
            load_checkpoint(
                self.online_net[0],
                self.pretrained,
                strict=False,
                logger=logger)
        else:
            self.online_net[0].init_weights()
        self.online_net[1].init_weights(init_linear='kaiming')  # projection
        for param_ol, param_tgt in zip(self.online_net.parameters(),
                                       self.target_net.parameters()):
            param_tgt.data.copy_(param_ol.data)
        # init the predictor in the head
        self.head.init_weights()

    @torch.no_grad()
    def _momentum_update(self):
        """
        Momentum update of the target network.
        """
        for param_ol, param_tgt in zip(self.online_net.parameters(),
                                       self.target_net.parameters()):
            param_tgt.data = param_tgt.data * self.momentum + \
                param_ol.data * (1. - self.momentum)

    def forward_train(self, img, **kwargs):
        assert isinstance(img, list)
        assert len(img) == 2
        for _img in img:
            assert _img.dim() == 4, \
                'Input must have 4 dims, got: {}'.format(_img.dim())

        img_v1 = img[0].contiguous()
        img_v2 = img[1].contiguous()

        # compute query features
        proj_online_v1 = self.online_net(img_v1)[0]
        proj_online_v2 = self.online_net(img_v2)[0]
        with torch.no_grad():
            proj_target_v1 = self.target_net(img_v1)[0].clone().detach()
            proj_target_v2 = self.target_net(img_v2)[0].clone().detach()

        loss = self.head(proj_online_v1, proj_target_v2)['loss'] + \
            self.head(proj_online_v2, proj_target_v1)['loss']
        self._momentum_update()
        return dict(loss=loss)

    def forward_test(self, img, **kwargs):
        pass

    def forward(self, img, mode='train', **kwargs):
        if mode == 'train':
            return self.forward_train(img, **kwargs)
        elif mode == 'test':
            return self.forward_test(img, **kwargs)
        elif mode == 'extract':
            return self.backbone(img)
        else:
            raise KeyError('No such mode: {}'.format(mode))



_base_ = ['configs/base.py']

model = dict(
    type='TextClassifier',
    backbone=dict(type='OCRRecMobileNetV3', scale=0.35, model_name='small'),
    head=dict(
        type='ClsHead',
        with_avg_pool=True,
        in_channels=200,
        num_classes=2,
    ),
    pretrained=
    'http://pai-vision-data-hz.oss-cn-zhangjiakou.aliyuncs.com/EasyCV/modelzoo/ocr/cls/ch_ppocr_mobile_v2.0_cls/best_accuracy.pth'
)

train_pipeline = [
    dict(type='RecAug', use_tia=False),
    dict(type='ClsResizeImg', img_shape=(3, 48, 192)),
    dict(type='MMToTensor'),
    dict(type='Collect', keys=['img', 'label'], meta_keys=['img_path'])
]

val_pipeline = [
    dict(type='ClsResizeImg', img_shape=(3, 48, 192)),
    dict(type='MMToTensor'),
    dict(type='Collect', keys=['img', 'label'], meta_keys=['img_path'])
]

test_pipeline = [
    dict(type='ClsResizeImg', img_shape=(3, 48, 192)),
    dict(type='MMToTensor'),
    dict(type='Collect', keys=['img'], meta_keys=['img_path'])
]

train_dataset = dict(
    type='OCRClsDataset',
    data_source=dict(
        type='OCRClsSource',
        label_file='ocr/direction/pai/label_file/test_direction.txt',
        data_dir='ocr/direction/pai/img/test',
        label_list=['0', '180'],
    ),
    pipeline=train_pipeline)

val_dataset = dict(
    type='OCRClsDataset',
    data_source=dict(
        type='OCRClsSource',
        label_file='ocr/direction/pai/label_file/test_direction.txt',
        data_dir='ocr/direction/pai/img/test',
        label_list=['0', '180'],
        test_mode=True),
    pipeline=val_pipeline)

data = dict(
    imgs_per_gpu=512, workers_per_gpu=8, train=train_dataset, val=val_dataset)

total_epochs = 100
optimizer = dict(type='Adam', lr=0.001, betas=(0.9, 0.999))

# learning policy
lr_config = dict(
    policy='CosineAnnealing',
    min_lr=1e-5,
    warmup='linear',
    warmup_iters=5,
    warmup_ratio=1e-4,
    warmup_by_epoch=True,
    by_epoch=False)

checkpoint_config = dict(interval=10)

log_config = dict(
    interval=10, hooks=[
        dict(type='TextLoggerHook'),
    ])

eval_config = dict(initial=True, interval=1, gpu_collect=False)
eval_pipelines = [
    dict(
        mode='test',
        data=data['val'],
        dist_eval=False,
        evaluators=[dict(type='ClsEvaluator', topk=(1, ))],
    )
]



# Copyright (c) Alibaba, Inc. and its affiliates.
from torch import nn

from easycv.utils.registry import build_from_cfg
from .registry import (ATTENTION, BACKBONES, FEEDFORWARD_NETWORK,
                       FUSION_LAYERS, HEADS, LOSSES, MIDDLE_ENCODERS, MODELS,
                       NECKS, POSITIONAL_ENCODING, TRANSFORMER,
                       TRANSFORMER_LAYER, TRANSFORMER_LAYER_SEQUENCE,
                       VOXEL_ENCODERS)


def build(cfg, registry, default_args=None):
    if isinstance(cfg, list):
        modules = [
            build_from_cfg(cfg_, registry, default_args) for cfg_ in cfg
        ]
        return nn.Sequential(*modules)
    else:
        return build_from_cfg(cfg, registry, default_args)


def build_backbone(cfg):
    return build(cfg, BACKBONES)


def build_neck(cfg):
    return build(cfg, NECKS)


def build_head(cfg):
    return build(cfg, HEADS)


def build_loss(cfg):
    return build(cfg, LOSSES)


def build_model(cfg):
    return build(cfg, MODELS)


def build_voxel_encoder(cfg):
    """Build voxel encoder."""
    return VOXEL_ENCODERS.build(cfg)


def build_middle_encoder(cfg):
    """Build middle level encoder."""
    return MIDDLE_ENCODERS.build(cfg)


def build_fusion_layer(cfg):
    """Build fusion layer."""
    return FUSION_LAYERS.build(cfg)


def build_transformer(cfg, default_args=None):
    """Builder for Transformer."""
    return build_from_cfg(cfg, TRANSFORMER, default_args)


def build_positional_encoding(cfg, default_args=None):
    """Builder for Position Encoding."""
    return build_from_cfg(cfg, POSITIONAL_ENCODING, default_args)


def build_attention(cfg, default_args=None):
    """Builder for attention."""
    return build_from_cfg(cfg, ATTENTION, default_args)


def build_feedforward_network(cfg, default_args=None):
    """Builder for feed-forward network (FFN)."""
    return build_from_cfg(cfg, FEEDFORWARD_NETWORK, default_args)


def build_transformer_layer(cfg, default_args=None):
    """Builder for transformer layer."""
    return build_from_cfg(cfg, TRANSFORMER_LAYER, default_args)


def build_transformer_layer_sequence(cfg, default_args=None):
    """Builder for transformer encoder and transformer decoder."""
    return build_from_cfg(cfg, TRANSFORMER_LAYER_SEQUENCE, default_args)



# Copyright (c) Alibaba, Inc. and its affiliates.
import torch
import torch.nn.functional as F
from torch import nn

from easycv.models import builder
from easycv.models.base import BaseModel
from easycv.models.builder import MODELS
from easycv.utils.checkpoint import get_checkpoint


# Modified from https://github.com/open-mmlab/mmaction2/blob/master/mmaction/models/recognizers/recognizer3d.py
@MODELS.register_module()
class Recognizer3D(BaseModel):
    """3D recognizer model.
    """

    def __init__(self,
                 backbone,
                 cls_head=None,
                 neck=None,
                 train_cfg=None,
                 test_cfg=None,
                 pretrained=None):
        super(Recognizer3D, self).__init__()
        self.backbone = builder.build_backbone(backbone)
        self.neck = builder.build_neck(neck) if neck else None
        self.cls_head = builder.build_head(cls_head) if cls_head else None

        self.train_cfg = train_cfg
        self.test_cfg = test_cfg
        self.pretrained = get_checkpoint(
            pretrained) if pretrained else pretrained
        self.activate_fn = nn.Softmax(dim=1)

        # aux_info is the list of tensor names beyond 'imgs' and 'label' which
        # will be used in train_step and val_step, data_batch should contain
        # these tensors
        self.aux_info = []
        if train_cfg is not None and 'aux_info' in train_cfg:
            self.aux_info = train_cfg['aux_info']
        # max_testing_views should be int
        self.max_testing_views = None
        if test_cfg is not None and 'max_testing_views' in test_cfg:
            self.max_testing_views = test_cfg['max_testing_views']
            assert isinstance(self.max_testing_views, int)

        if test_cfg is not None and 'feature_extraction' in test_cfg:
            self.feature_extraction = test_cfg['feature_extraction']
        else:
            self.feature_extraction = False

        # mini-batch blending, e.g. mixup, cutmix, etc.
        # self.blending = None
        # if train_cfg is not None and 'blending' in train_cfg:
        #     from mmcv.utils import build_from_cfg
        #     from mmaction.datasets.builder import BLENDINGS
        #     self.blending = build_from_cfg(train_cfg['blending'], BLENDINGS)

        self.init_weights()

    @property
    def with_neck(self):
        """bool: whether the recognizer has a neck"""
        return hasattr(self, 'neck') and self.neck is not None

    @property
    def with_cls_head(self):
        """bool: whether the recognizer has a cls_head"""
        return hasattr(self, 'cls_head') and self.cls_head is not None

    def init_weights(self):
        """Initialize the model network weights."""
        if isinstance(self.pretrained, str):
            self.backbone.init_weights(pretrained=self.pretrained)
        else:
            self.backbone.init_weights()

        if self.with_cls_head and hasattr(self.cls_head, 'init_weights'):
            self.cls_head.init_weights()
        if self.with_neck and hasattr(self.neck, 'init_weights'):
            self.neck.init_weights()

    def average_clip(self, cls_score, num_segs=1):
        """Averaging class score over multiple clips.
        Using different averaging types ('score' or 'prob' or None,
        which defined in test_cfg) to computed the final averaged
        class score. Only called in test mode.
        Args:
            cls_score (torch.Tensor): Class score to be averaged.
            num_segs (int): Number of clips for each input sample.
        Returns:
            torch.Tensor: Averaged class score.
        """
        if 'average_clips' not in self.test_cfg.keys():
            raise KeyError('"average_clips" must defined in test_cfg\'s keys')

        average_clips = self.test_cfg['average_clips']
        if average_clips not in ['score', 'prob', None]:
            raise ValueError(f'{average_clips} is not supported. '
                             f'Currently supported ones are '
                             f'["score", "prob", None]')

        if average_clips is None:
            return cls_score

        batch_size = cls_score.shape[0]
        cls_score = cls_score.view(batch_size // num_segs, num_segs, -1)

        if average_clips == 'prob':
            cls_score = F.softmax(cls_score, dim=2).mean(dim=1)
        elif average_clips == 'score':
            cls_score = cls_score.mean(dim=1)

        return cls_score

    def extract_feat(self, imgs):
        """Extract features through a backbone.
        Args:
            imgs (torch.Tensor): The input images.
        Returns:
            torch.tensor: The extracted features.
        """

        x = self.backbone(imgs)
        return x

    def forward_train(self, imgs, label, **kwargs):
        """Defines the computation performed at every call when training."""

        assert self.with_cls_head
        imgs = imgs.reshape((-1, ) + imgs.shape[2:])
        losses = dict()

        x = self.extract_feat(imgs)
        if self.with_neck:
            x, loss_aux = self.neck(x, label.squeeze())
            losses.update(loss_aux)

        cls_score = self.cls_head(x)
        gt_labels = label.squeeze()
        loss_cls = self.cls_head.loss(cls_score, gt_labels)
        losses.update(loss_cls)

        return losses

    def _do_test(self, imgs):
        """Defines the computation performed at every call when evaluation,
        testing and gradcam."""
        batches = imgs.shape[0]
        num_segs = imgs.shape[1]
        imgs = imgs.reshape((-1, ) + imgs.shape[2:])

        if self.max_testing_views is not None:
            total_views = imgs.shape[0]
            assert num_segs == total_views, (
                'max_testing_views is only compatible '
                'with batch_size == 1')
            view_ptr = 0
            feats = []
            while view_ptr < total_views:
                batch_imgs = imgs[view_ptr:view_ptr + self.max_testing_views]
                x = self.extract_feat(batch_imgs)
                if self.with_neck:
                    x, _ = self.neck(x)
                feats.append(x)
                view_ptr += self.max_testing_views
            # should consider the case that feat is a tuple
            if isinstance(feats[0], tuple):
                len_tuple = len(feats[0])
                feat = [
                    torch.cat([x[i] for x in feats]) for i in range(len_tuple)
                ]
                feat = tuple(feat)
            else:
                feat = torch.cat(feats)
        else:
            feat = self.extract_feat(imgs)
            if self.with_neck:
                feat, _ = self.neck(feat)

        if self.feature_extraction:
            # perform spatio-temporal pooling
            avg_pool = nn.AdaptiveAvgPool3d(1)
            if isinstance(feat, tuple):
                feat = [avg_pool(x) for x in feat]
                # concat them
                feat = torch.cat(feat, axis=1)
            else:
                feat = avg_pool(feat)
            # squeeze dimensions
            feat = feat.reshape((batches, num_segs, -1))
            # temporal average pooling
            feat = feat.mean(axis=1)
            return feat

        # should have cls_head if not extracting features
        assert self.with_cls_head
        cls_score = self.cls_head(feat)
        cls_score = self.average_clip(cls_score, num_segs)
        return cls_score

    def forward_test(self, imgs, label=None, **kwargs):
        """Defines the computation performed at every call when evaluation and
        testing."""
        cls_score = self._do_test(imgs)
        if label is not None:
            return dict(neck=cls_score.cpu(), label=label.cpu())
        else:
            result = {}
            result['prob'] = self.activate_fn(cls_score.cpu())
            if 'img_metas' in kwargs and 'filename' in kwargs['img_metas'][0]:
                result['filename'] = [kwargs['img_metas'][0]['filename']]
            # result['class'] = torch.argmax(result['prob'])
            # print(result['class'])
            return result

    def train_step(self, data, optimizer):
        """The iteration step during training.

        This method defines an iteration step during training, except for the
        back propagation and optimizer updating, which are done in an optimizer
        hook. Note that in some complicated cases or models, the whole process
        including back propagation and optimizer updating is also defined in
        this method, such as GAN.

        Args:
            data (dict): The output of dataloader.
            optimizer (:obj:`torch.optim.Optimizer` | dict): The optimizer of
                runner is passed to ``train_step()``. This argument is unused
                and reserved.

        Returns:
            dict: It should contain at least 3 keys: ``loss``, ``log_vars``, \
                ``num_samples``.

                - ``loss`` is a tensor for back propagation, which can be a \
                weighted sum of multiple losses.
                - ``log_vars`` contains all the variables to be sent to the \
                logger.
                - ``num_samples`` indicates the batch size (when the model is \
                DDP, it means the batch size on each GPU), which is used for \
                averaging the logs.
        """
        losses = self(**data, mode='train')
        loss, log_vars = self._parse_losses(losses)

        if type(data['imgs']) == list:
            num_samples = len(data['imgs'][0])
        else:
            num_samples = len(data['imgs'].data)

        return dict(loss=loss, log_vars=log_vars, num_samples=num_samples)



# Copyright (c) OpenMMLab. All rights reserved.
# Copyright (c) Alibaba, Inc. and its affiliates.
from abc import ABCMeta, abstractmethod


class BaseBBoxCoder(metaclass=ABCMeta):
    """Base bounding box coder."""

    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def encode(self, bboxes, gt_bboxes):
        """Encode deltas between bboxes and ground truth boxes."""

    @abstractmethod
    def decode(self, bboxes, bboxes_pred):
        """Decode the predicted bboxes according to prediction and base
        boxes."""



# Copyright (c) Alibaba, Inc. and its affiliates.
# flake8: noqa
# isort:skip_file
import os
os.environ['SETUPTOOLS_USE_DISTUTILS'] = 'stdlib'

from .version import __version__, short_version

__all__ = ['__version__', 'short_version']



# Copyright (c) Alibaba, Inc. and its affiliates.
import torch
import torch.nn as nn
import torch.nn.functional as F

from easycv.framework.errors import KeyError, ValueError
from easycv.utils.checkpoint import load_checkpoint
from easycv.utils.logger import get_root_logger
from easycv.utils.preprocess_function import gaussianBlur, randomGrayScale
from .. import builder
from ..base import BaseModel
from ..registry import MODELS


@MODELS.register_module
class MoBY(BaseModel):
    '''MoBY.
        Part of the code is borrowed from:
        https://github.com/SwinTransformer/Transformer-SSL/blob/main/models/moby.py.
    '''

    def __init__(self,
                 backbone,
                 train_preprocess=[],
                 neck=None,
                 head=None,
                 pretrained=None,
                 queue_len=4096,
                 contrast_temperature=0.2,
                 momentum=0.99,
                 online_drop_path_rate=0.2,
                 target_drop_path_rate=0.0,
                 **kwargs):
        """ Init Moby

        Args:
            backbone: backbone config to build vision backbone
            train_preprocess: [gaussBlur, mixUp, solarize]
            neck: neck config to build Moby Neck
            head: head config to build Moby Neck
            pretrained: pretrained weight for backbone
            queue_len :  moby queue length
            contrast_temperature : contrastive_loss temperature
            momentum : ema target weights momentum
            online_drop_path_rate: for transformer based backbone, set online model drop_path_rate
            target_drop_path_rate: for transformer based backbone, set target model drop_path_rate
        """

        super(MoBY, self).__init__()

        self.pretrained = pretrained

        self.preprocess_key_map = {
            'randomGrayScale': randomGrayScale,
            'gaussianBlur': gaussianBlur
        }
        self.train_preprocess = [
            self.preprocess_key_map[i] for i in train_preprocess
        ]

        # build model
        if backbone.get('drop_path_rate', None) is not None:
            backbone['drop_path_rate'] = online_drop_path_rate
        self.encoder_q = builder.build_backbone(backbone)

        if backbone.get('drop_path_rate', None) is not None:
            backbone['drop_path_rate'] = target_drop_path_rate
        self.encoder_k = builder.build_backbone(backbone)

        self.backbone = self.encoder_q

        self.projector_q = builder.build_neck(neck)
        self.projector_k = builder.build_neck(neck)
        self.predictor = builder.build_neck(head)

        # copy param, set stop_grad
        for param_q, param_k in zip(self.encoder_q.parameters(),
                                    self.encoder_k.parameters()):
            param_k.data.copy_(param_q.data)  # initialize
            param_k.requires_grad = False  # not update by gradient
        for param_q, param_k in zip(self.projector_q.parameters(),
                                    self.projector_k.parameters()):
            param_k.data.copy_(param_q.data)
            param_k.requires_grad = False

        # convert bn to sync bn
        try:
            nn.SyncBatchNorm.convert_sync_batchnorm(self.encoder_q)
            nn.SyncBatchNorm.convert_sync_batchnorm(self.encoder_k)
        except Exception as e:
            print('Convert encode BN to syncBN failed for MoBY backbone %s' %
                  (str(type(self.encoder_q))))

        nn.SyncBatchNorm.convert_sync_batchnorm(self.projector_q)
        nn.SyncBatchNorm.convert_sync_batchnorm(self.projector_k)
        nn.SyncBatchNorm.convert_sync_batchnorm(self.predictor)

        # set parameters
        self.init_weights()
        self.queue_len = queue_len
        self.momentum = momentum
        self.contrast_temperature = contrast_temperature
        self.feat_dim = head.get('out_channels', 256)
        assert neck.get('out_channels', 256) == head.get(
            'out_channels',
            256), 'MoBY head and neck should set same output dim'

        # create the queue
        self.register_buffer('queue1',
                             torch.randn(self.feat_dim, self.queue_len))
        self.register_buffer('queue2',
                             torch.randn(self.feat_dim, self.queue_len))
        self.queue1 = F.normalize(self.queue1, dim=0)
        self.queue2 = F.normalize(self.queue2, dim=0)
        self.register_buffer('queue_ptr', torch.zeros(1, dtype=torch.long))

        self.avg_pool = nn.AdaptiveAvgPool2d((1, 1))

    def init_weights(self):
        if isinstance(self.pretrained, str):
            logger = get_root_logger()
            load_checkpoint(
                self.encoder_q, self.pretrained, strict=False, logger=logger)
        else:
            self.encoder_q.init_weights()
        for param_q, param_k in zip(self.encoder_q.parameters(),
                                    self.encoder_k.parameters()):
            param_k.data.copy_(param_q.data)

    def forward_backbone(self, img):
        feature_list = self.backbone(img)
        return feature_list

    @torch.no_grad()
    def _momentum_update_key_encoder(self):
        """
        Momentum update of the key encoder
        """
        # need a scheduler
        _contrast_momentum = self.momentum
        for param_q, param_k in zip(self.encoder_q.parameters(),
                                    self.encoder_k.parameters()):
            param_k.data = param_k.data * _contrast_momentum + param_q.data * (
                1. - _contrast_momentum)

        for param_q, param_k in zip(self.projector_q.parameters(),
                                    self.projector_k.parameters()):
            param_k.data = param_k.data * _contrast_momentum + param_q.data * (
                1. - _contrast_momentum)

    @torch.no_grad()
    def _dequeue_and_enqueue(self, keys1, keys2):
        # gather keys before updating queue
        keys1 = concat_all_gather(keys1)
        keys2 = concat_all_gather(keys2)

        batch_size = keys1.shape[0]

        ptr = int(self.queue_ptr)
        assert self.queue_len % batch_size == 0  # for simplicity

        # replace the keys at ptr (dequeue and enqueue)
        self.queue1[:, ptr:ptr + batch_size] = keys1.transpose(0, 1)
        self.queue2[:, ptr:ptr + batch_size] = keys2.transpose(0, 1)
        ptr = (ptr + batch_size) % self.queue_len  # move pointer
        self.queue_ptr[0] = ptr

    def contrastive_loss(self, q, k, queue):

        # positive logits: Nx1
        l_pos = torch.einsum('nc,nc->n', [q, k]).unsqueeze(-1)
        # negative logits: NxK
        l_neg = torch.einsum('nc,ck->nk', [q, queue.clone().detach()])

        # logits: Nx(1+K)
        logits = torch.cat([l_pos, l_neg], dim=1)

        # apply temperature
        logits /= self.contrast_temperature

        # labels: positive key indicators
        labels = torch.zeros(logits.shape[0], dtype=torch.long).cuda()

        return F.cross_entropy(logits, labels)

    def forward_train(self, img, **kwargs):
        assert isinstance(img, list)
        assert len(img) == 2
        for _img in img:
            assert _img.dim() == 4, \
                'Input must have 4 dims, got: {}'.format(_img.dim())

        im_q = img[0].contiguous()
        im_k = img[1].contiguous()

        for preprocess in self.train_preprocess:
            im_q = preprocess(im_q)
            im_k = preprocess(im_k)

        # compute query features
        feat_1 = self.encoder_q(im_q)  # queries: NxC
        proj_1 = self.projector_q(feat_1)
        pred_1 = self.predictor(proj_1)[0]
        pred_1 = F.normalize(pred_1, dim=1)

        feat_2 = self.encoder_q(im_k)
        proj_2 = self.projector_q(feat_2)
        pred_2 = self.predictor(proj_2)[0]
        pred_2 = F.normalize(pred_2, dim=1)

        # compute key features
        with torch.no_grad():  # no gradient to keys
            self._momentum_update_key_encoder()  # update the key encoder

            feat_1_ng = self.encoder_k(im_q)  # keys: NxC
            proj_1_ng = self.projector_k(feat_1_ng)[0]
            proj_1_ng = F.normalize(proj_1_ng, dim=1)

            feat_2_ng = self.encoder_k(im_k)
            proj_2_ng = self.projector_k(feat_2_ng)[0]
            proj_2_ng = F.normalize(proj_2_ng, dim=1)

        # compute loss
        losses = dict()

        losses['loss'] = self.contrastive_loss(pred_1, proj_2_ng, self.queue2) \
            + self.contrastive_loss(pred_2, proj_1_ng, self.queue1)

        self._dequeue_and_enqueue(proj_1_ng, proj_2_ng)

        return losses

    def forward_test(self, img, **kwargs):
        pass

    def forward_feature(self, img, **kwargs):
        """Forward backbone

        Returns:
            x (torch.Tensor): feature tensor
        """
        return_dict = {}
        x = self.backbone(img)
        return_dict['backbone'] = x

        if hasattr(self, 'neck') and self.neck is not None:
            feature = self.neck([self.avg_pool(i) for i in x])[0]
        else:
            feature = self.avg_pool(x[-1])
        return_dict['neck'] = feature

        return return_dict

    def forward(self,
                img,
                gt_label=None,
                mode='train',
                extract_list=['neck'],
                **kwargs):
        if mode == 'train':
            return self.forward_train(img, **kwargs)
        elif mode == 'test':
            return self.forward_test(img, **kwargs)

        elif mode == 'extract':
            rd = self.forward_feature(img)
            rv = {}

            for name in extract_list:
                if name in rd.keys():
                    rv[name] = rd[name]
                else:
                    raise ValueError(
                        'Extract %s is not support in classification models' %
                        name)
            if gt_label is not None:
                rv['gt_labels'] = gt_label.cpu()
            return rv
        else:
            raise KeyError('No such mode: {}'.format(mode))


# utils
@torch.no_grad()
def concat_all_gather(tensor):
    """
    Performs all_gather operation on the provided tensors.
    *** Warning ***: torch.distributed.all_gather has no gradient.
    """
    tensors_gather = [
        torch.ones_like(tensor)
        for _ in range(torch.distributed.get_world_size())
    ]
    torch.distributed.all_gather(tensors_gather, tensor, async_op=False)

    output = torch.cat(tensors_gather, dim=0)
    return output
